import torch
import dgl
import torch.nn as nn
import torch.nn.functional as F
import dgl.nn as dglnn
import dgl.function as fn
import math

from . import BaseModel, register_model
from ..layers.EmbedLayer import HeteroEmbedLayer
from ..utils import get_nodes_dict


@register_model('HGT')
class HGT(BaseModel):
    @classmethod
    def build_model_from_args(cls, args, hg):
        node_dict = {}
        edge_dict = {}

        for ntype in hg.ntypes:
            node_dict[ntype] = len(node_dict)
        for etype in hg.etypes:
            edge_dict[etype] = len(edge_dict)
            hg.edges[etype].data['id'] = torch.ones(hg.number_of_edges(etype), dtype=torch.long).to(args.device) * edge_dict[etype]
        in_dim = hg.nodes['paper'].data['feat'].shape[1] + 1

        return cls(in_dim, args.hidden_dim, args.num_heads, len(hg.etypes), len(hg.ntypes), args.n_layers, args.num_heads)

    def __init__(self, in_size, out_size, n_heads, n_etypes, n_ntypes, n_layers, n_classes, dropout=0.2):
        super().__init__()

        self.adapt_ws = TypedLinear(n_ntypes, in_size, out_size)
        self.convs = nn.ModuleList()
        self.norms = nn.ModuleList()
        for i in range(n_layers):
            self.convs.append(HGTConv(out_size, out_size, n_heads, n_etypes, n_ntypes, dropout))
            self.norms.append(TypedLayerNorm(n_ntypes, out_size))
        self.dropout = nn.Dropout(dropout)
        self.classifier = nn.Linear(out_size, n_classes)

    def forward(self, g, x):
        g = g
        x = x
        h = self.dropout(torch.tanh(self.adapt_ws(x, g.ndata['ntype'])))
        for i, (conv, norm) in enumerate(zip(self.convs, self.norms)):
            h = norm(conv(g, h), g.ndata['ntype'])
        return self.classifier(h)


class RelTemporalEncoding(nn.Module):
    '''
        Implement the Temporal Encoding (Sinusoid) function.
    '''

    def __init__(self, n_hid, max_len=240):
        super(RelTemporalEncoding, self).__init__()

        position = torch.arange(0., max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, n_hid, 2) * -(math.log(10000.0) / n_hid))
        self.emb = nn.Embedding(max_len, n_hid)
        self.emb.weight.data[:, 0::2] = torch.sin(position * div_term) / (n_hid ** 0.5)
        self.emb.weight.data[:, 1::2] = torch.cos(position * div_term) / (n_hid ** 0.5)
        self.emb.requires_grad = False
        self.lin = nn.Linear(n_hid, n_hid)

    def forward(self, x, t):
        return x + self.lin(self.emb(t))


class TypedLinear(nn.Module):
    def __init__(self, num_types, in_size, out_size, bias=True):
        super().__init__()
        self.bias = bias
        self.W = nn.Parameter(torch.zeros(num_types, out_size, in_size))
        if bias:
            self.b = nn.Parameter(torch.zeros(num_types, out_size))

        nn.init.xavier_uniform_(self.W)

    def forward(self, x, types):
        """
        x : float[b, ..., in_size]
        types : int64[b]
        """
        return iaddmm(x, types, self.W, self.b) if self.bias else imm(x, types, self.W)


class TypedLayerNorm(nn.Module):
    def __init__(self, num_types, size):
        super().__init__()

        self.W = nn.Parameter(torch.ones(num_types, size))
        self.b = nn.Parameter(torch.zeros(num_types, size))

    def forward(self, x, types):
        """
        x : float[b, ..., size]
        types : int64[b]
        """
        return typed_layer_norm(x, types, self.W, self.b)


class HGTMessage(nn.Module):
    def __init__(self, in_size, out_size, n_heads, n_etypes, n_ntypes):
        super().__init__()

        assert out_size % n_heads == 0
        self.d_k = out_size // n_heads
        self.n_heads = n_heads

        self.rte = RelTemporalEncoding(in_size)
        self.K = TypedLinear(n_ntypes, in_size, out_size)
        self.V = TypedLinear(n_ntypes, in_size, out_size)
        self.W_att = TypedLinear(n_etypes, self.d_k, self.d_k, bias=False)
        self.W_msg = TypedLinear(n_etypes, self.d_k, self.d_k, bias=False)
        self.mu = nn.Parameter(torch.ones(n_etypes, n_heads))

    def forward(self, edges):
        phi = edges.data['_TYPE']  # int64[E]
        tau_s = edges.src['_TYPE']  # int64[E]
        #tau_t = edges.dst['_TYPE']  # int64[E]
        h_s = edges.src['h']  # float32[E, in_size]
        Q_t = edges.dst['Q']  # float32[E, n_heads, d_k]
        # dT = edges.data['dt']  # int64[E]

        # Step 1. compute relative time encoding
        # NOTE: although the figure in the paper writes something \hat{H}^{(l-1)}[s_1] which looks as
        # if it only depends on node s_1, it actually also depends on the target node t because the
        # relative time difference can be different if t differs.  This is confirmed in the original
        # HGT implementation where h_hat_s is computed inside the message function.
        # h_hat_s = self.rte(h_s, dT)  # float32[E, in_size]
        h_hat_s = h_s
        # Step 2. compute source keys, and source messages
        # Target queries are computed in HGTConv to save memory
        K_s = self.K(h_hat_s, tau_s).view(-1, self.n_heads, self.d_k)  # float32[E, n_heads, d_k]
        V_s = self.V(h_hat_s, tau_s).view(-1, self.n_heads, self.d_k)  # float32[E, n_heads, d_k]
        M = self.W_msg(V_s, phi)  # float32[E, n_heads, d_k]

        # Step 3. compute attention logits over edges ATT-head.  It will be normalized with edge_softmax
        # later.
        # NOTE: the formulation K^i W_{ATT} Q^i is wrong.  We instead need the attention score computed on
        # the edges.  This is confirmed in the originial HGT implementation where the score is computed
        # with a sum reduction rather than a matrix multiplication.
        att_head_k = self.W_att(K_s, phi)
        att_head = (Q_t * att_head_k).sum(-1)  # float32[E, n_heads]
        att_head = att_head * self.mu[phi] / (self.d_k ** 0.5)  # float32[E, n_heads]

        return {'att_head': att_head, 'M': M}


class HGTConv(nn.Module):
    def __init__(self, in_size, out_size, n_heads, n_etypes, n_ntypes, dropout=0.2):
        super().__init__()

        # assert in_size == out_size
        self.out_size = out_size
        self.d_k = out_size // n_heads
        self.n_heads = n_heads

        self.apply_edges = HGTMessage(in_size, out_size, n_heads, n_etypes, n_ntypes)
        self.A = TypedLinear(n_ntypes, out_size, out_size)
        self.Q = TypedLinear(n_ntypes, in_size, out_size)
        self.dropout = nn.Dropout(dropout)
        self.alpha = nn.Parameter(torch.ones(n_ntypes))

    def forward(self, g, h):
        with g.local_scope():
            g.ndata['h'] = h

            # Step 1-3
            g.ndata['Q'] = self.Q(h, g.ndata['_TYPE']).view(-1, self.n_heads, self.d_k)
            g.apply_edges(self.apply_edges)

            # Step 4. compute attention score and do weighted average of messages
            g.edata['att_score'] = dglnn.edge_softmax(g, g.edata['att_head'])
            g.edata['M'] = g.edata['M'] * g.edata['att_score'][..., None]
            g.update_all(fn.copy_e('M', 'M'), fn.sum('M', 'h_new'))
            h_new = g.ndata['h_new'].view(-1, self.out_size)

            # Step 5. update.
            # NOTE: the official implementation uses learnable gating rather than residual connections to combine
            # the aggregated messages and the node representation itself.
            # h_new = self.dropout(self.A(F.gelu(h_new), g.ndata['ntype']))
            # alpha = F.sigmoid(self.alpha[g.ndata['ntype']])[:, None]
            # h_out = h_new * alpha + h * (1 - alpha)
            return h_new


class IAddMM(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, types, W, b):
        """
        x : float[B, ..., in_size]
        types : (0..n_types)[B]
        W : float[n_types, out_size, in_size]
        b : float[n_types, out_size]
        return : float[B, ..., out_size] = xW[types] + b[types]
        """
        B = x.shape[0]
        mid_size = x.shape[1:-1]
        mid_dims = len(mid_size)
        in_size = x.shape[-1]
        out_size = W.shape[1]

        y = torch.zeros(B, *mid_size, out_size, device=x.device)

        for t in range(types.max().item() + 1):
            t_mask = (types == t)
            if not t_mask.any():
                continue
            W_t = W[t]
            b_t = b[t]
            x_t = x[t_mask].unsqueeze(-1)
            y_t = (W_t @ x_t).squeeze(-1) + b_t
            y[t_mask] = y_t

        ctx.save_for_backward(x, types, W, b)
        return y

    @staticmethod
    def backward(ctx, grad_y):
        x, types, W, b = ctx.saved_tensors
        grad_x = torch.zeros_like(x)
        grad_W = torch.zeros_like(W)
        grad_b = torch.zeros_like(b)

        for t in range(types.max().item() + 1):
            t_mask = (types == t)
            if not t_mask.any():
                continue
            grad_y_t = grad_y[t_mask]
            grad_x[t_mask] = grad_y_t @ W[t]
            grad_W[t] = grad_y_t.view(-1, grad_y_t.shape[-1]).T @ x[t_mask].view(-1, x.shape[-1])
            grad_b[t] = grad_y_t.view(-1, b.shape[-1]).sum(0)
        return grad_x, None, grad_W, grad_b


def iaddmm(x, types, W, b):
    return IAddMM.apply(x, types, W, b)


class IMM(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, types, W):
        """
        x : float[B, ..., in_size]
        types : (0..n_types)[B]
        W : float[n_types, out_size, in_size]
        return : float[B, ..., out_size] = xW[types] + b[types]
        """
        B = x.shape[0]
        mid_size = x.shape[1:-1]
        mid_dims = len(mid_size)
        in_size = x.shape[-1]
        out_size = W.shape[1]

        y = torch.zeros(B, *mid_size, out_size, device=x.device)

        for t in range(types.max().item() + 1):
            t_mask = (types == t)
            if not t_mask.any():
                continue
            W_t = W[t]
            x_t = x[t_mask].unsqueeze(-1)
            y_t = (W_t @ x_t).squeeze(-1)
            y[t_mask] = y_t

        ctx.save_for_backward(x, types, W)
        return y

    @staticmethod
    def backward(ctx, grad_y):
        x, types, W = ctx.saved_tensors
        grad_x = torch.zeros_like(x)
        grad_W = torch.zeros_like(W)

        for t in range(types.max().item() + 1):
            t_mask = (types == t)
            if not t_mask.any():
                continue
            grad_y_t = grad_y[t_mask]
            grad_x[t_mask] = grad_y_t @ W[t]
            grad_W[t] = grad_y_t.view(-1, grad_y_t.shape[-1]).T @ x[t_mask].view(-1, x.shape[-1])
        return grad_x, None, grad_W


def imm(x, types, W):
    return IMM.apply(x, types, W)


class TLayerNorm(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, types, w, b):
        ctx.save_for_backward(x, types, w, b)
        x = (x - x.mean(-1, keepdim=True)) / (x.var(-1, keepdim=True, unbiased=False) + 1e-5).sqrt()
        w = w[types][(slice(None),) + (None,) * (x.dim() - 2)]
        b = b[types][(slice(None),) + (None,) * (x.dim() - 2)]
        return x * w + b

    @staticmethod
    def backward(ctx, grad_y):
        x, types, w, b = ctx.saved_tensors
        types_expanded = types[(slice(None),) + (None,) * (x.dim() - 1)].expand_as(grad_y)

        delta = x - x.mean(-1, keepdim=True)
        v = (x.var(-1, keepdim=True, unbiased=False) + 1e-5).sqrt()
        x_norm = delta / v
        n = x.shape[-1]

        grad_w = torch.zeros(w.shape[0], *x.shape[1:], device=w.device)
        grad_w.scatter_add_(0, types_expanded, grad_y * x_norm)
        grad_w = torch.einsum('i...j->ij', grad_w)

        grad_b = torch.zeros(b.shape[0], *x.shape[1:], device=b.device)
        grad_b.scatter_add_(0, types_expanded, grad_y)
        grad_b = torch.einsum('i...j->ij', grad_b)

        w_t = w[types][(slice(None),) + (None,) * (x.dim() - 2)].expand_as(grad_y)
        grad_x = grad_y * w_t / v
        grad_x += -1 / n / v * torch.einsum('...ij,...ij->...i', grad_y, w_t)[..., None]
        grad_x += -1 / n / (v ** 3) * delta * torch.einsum('...ij,...ij,...ij->...i', grad_y, delta, w_t)[..., None]

        return grad_x, None, grad_w, grad_b


def typed_layer_norm(x, types, w, b):
    return TLayerNorm.apply(x, types, w, b)


import dgl
import torch as th
import torch.nn as nn
import torch.nn.functional as F
import dgl.nn as dglnn
from . import BaseModel, register_model


@register_model('RGAT')
class RGAT(BaseModel):
    @classmethod
    def build_model_from_args(cls, args, hg):
        return cls(in_dim=args.hidden_dim,
                   out_dim=args.hidden_dim,
                   h_dim=args.out_dim,
                   etypes=hg.etypes,
                   num_heads=args.num_heads,
                   dropout=args.dropout)

    def __init__(self, in_dim, out_dim, h_dim, etypes, num_heads, dropout):
        super(RGAT, self).__init__()
        self.rel_names = etypes
        self.layers = nn.ModuleList()
        # input 2 hidden
        self.layers.append(RGATLayer(
            in_dim, h_dim, num_heads, self.rel_names, activation=F.relu, dropout=dropout))
        self.layers.append(RGATLayer(
            h_dim, out_dim, num_heads, self.rel_names, activation=None))
        return

    def forward(self, hg, h_dict=None):
        if hasattr(hg, 'ntypes'):
            # full graph training,
            for layer in self.layers:
                h_dict = layer(hg, h_dict)
        else:
            # minibatch training, block
            for layer, block in zip(self.layers, hg):
                h_dict = layer(block, h_dict)
        return h_dict


class RGATLayer(nn.Module):

    def __init__(self,
                 in_feat,
                 out_feat,
                 num_heads,
                 rel_names,
                 activation=None,
                 dropout=0.0,
                 bias=True,):
        super(RGATLayer, self).__init__()
        self.in_feat = in_feat
        self.out_feat = out_feat
        self.num_heads= num_heads
        self.conv = dglnn.HeteroGraphConv({
            rel: dgl.nn.pytorch.GATConv(in_feat, out_feat, num_heads=num_heads, bias=bias, allow_zero_in_degree=True)
            for rel in rel_names
        })

    def forward(self, g, h_dict):
        h_dict = self.conv(g, h_dict)
        out_put = {}
        for n_type, h in h_dict.items():
            out_put[n_type] = h.squeeze()
        return out_put


import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import init
import random
import numpy as np
import dgl.multiprocessing as mp
from dgl.multiprocessing import Queue


def init_emb2pos_index(walk_length, window_size, batch_size):
    ''' select embedding of positive nodes from a batch of node embeddings

    Return
    ------
    index_emb_posu torch.LongTensor : the indices of u_embeddings
    index_emb_posv torch.LongTensor : the indices of v_embeddings
    Usage
    -----
    # emb_u.shape: [batch_size * walk_length, dim]
    batch_emb2posu = torch.index_select(emb_u, 0, index_emb_posu)
    '''
    idx_list_u = []
    idx_list_v = []
    for b in range(batch_size):
        for i in range(walk_length):
            for j in range(i - window_size, i):
                if j >= 0:
                    idx_list_u.append(j + b * walk_length)
                    idx_list_v.append(i + b * walk_length)
            for j in range(i + 1, i + 1 + window_size):
                if j < walk_length:
                    idx_list_u.append(j + b * walk_length)
                    idx_list_v.append(i + b * walk_length)

    # [num_pos * batch_size]
    index_emb_posu = torch.LongTensor(idx_list_u)
    index_emb_posv = torch.LongTensor(idx_list_v)

    return index_emb_posu, index_emb_posv


def init_emb2neg_index(walk_length, window_size, negative, batch_size):
    '''select embedding of negative nodes from a batch of node embeddings
    for fast negative sampling

    Return
    ------
    index_emb_negu torch.LongTensor : the indices of u_embeddings
    index_emb_negv torch.LongTensor : the indices of v_embeddings
    Usage
    -----
    # emb_u.shape: [batch_size * walk_length, dim]
    batch_emb2negu = torch.index_select(emb_u, 0, index_emb_negu)
    '''
    idx_list_u = []
    for b in range(batch_size):
        for i in range(walk_length):
            for j in range(i - window_size, i):
                if j >= 0:
                    idx_list_u += [i + b * walk_length] * negative
            for j in range(i + 1, i + 1 + window_size):
                if j < walk_length:
                    idx_list_u += [i + b * walk_length] * negative

    idx_list_v = list(range(batch_size * walk_length)) \
                 * negative * window_size * 2
    random.shuffle(idx_list_v)
    idx_list_v = idx_list_v[:len(idx_list_u)]

    # [bs * walk_length * negative]
    index_emb_negu = torch.LongTensor(idx_list_u)
    index_emb_negv = torch.LongTensor(idx_list_v)

    return index_emb_negu, index_emb_negv


def init_weight(walk_length, window_size, batch_size):
    ''' init context weight '''
    weight = []
    for b in range(batch_size):
        for i in range(walk_length):
            for j in range(i - window_size, i):
                if j >= 0:
                    weight.append(1. - float(i - j - 1) / float(window_size))
            for j in range(i + 1, i + 1 + window_size):
                if j < walk_length:
                    weight.append(1. - float(j - i - 1) / float(window_size))

    # [num_pos * batch_size]
    return torch.Tensor(weight).unsqueeze(1)


def init_empty_grad(emb_dimension, walk_length, batch_size):
    """ initialize gradient matrix """
    grad_u = torch.zeros((batch_size * walk_length, emb_dimension))
    grad_v = torch.zeros((batch_size * walk_length, emb_dimension))

    return grad_u, grad_v


def adam(grad, state_sum, nodes, lr, device, only_gpu):
    """ calculate gradients according to adam """
    grad_sum = (grad * grad).mean(1)
    if not only_gpu:
        grad_sum = grad_sum.cpu()
    state_sum.index_add_(0, nodes, grad_sum)  # cpu
    std = state_sum[nodes].to(device)  # gpu
    std_values = std.sqrt_().add_(1e-10).unsqueeze(1)
    grad = (lr * grad / std_values)  # gpu

    return grad


def async_update(num_threads, model, queue):
    """ asynchronous embedding update """
    torch.set_num_threads(num_threads)
    while True:
        (grad_u, grad_v, grad_v_neg, nodes, neg_nodes) = queue.get()
        if grad_u is None:
            return
        with torch.no_grad():
            model.u_embeddings.weight.data.index_add_(0, nodes.view(-1), grad_u)
            model.v_embeddings.weight.data.index_add_(0, nodes.view(-1), grad_v)
            if neg_nodes is not None:
                model.v_embeddings.weight.data.index_add_(0, neg_nodes.view(-1), grad_v_neg)


class SkipGramModel(nn.Module):
    """ Negative sampling based skip-gram """

    def __init__(self,
                 emb_size,
                 emb_dimension,
                 walk_length,
                 window_size,
                 batch_size,
                 only_cpu,
                 only_gpu,
                 mix,
                 neg_weight,
                 negative,
                 lr,
                 lap_norm,
                 fast_neg,
                 record_loss,
                 norm,
                 use_context_weight,
                 async_update,
                 num_threads,
                 ):
        """ initialize embedding on CPU
        Paremeters
        ----------
        emb_size int : number of nodes
        emb_dimension int : embedding dimension
        walk_length int : number of nodes in a sequence
        window_size int : context window size
        batch_size int : number of node sequences in each batch
        only_cpu bool : training with CPU
        only_gpu bool : training with GPU
        mix bool : mixed training with CPU and GPU
        negative int : negative samples for each positve node pair
        neg_weight float : negative weight
        lr float : initial learning rate
        lap_norm float : weight of laplacian normalization
        fast_neg bool : do negative sampling inside a batch
        record_loss bool : print the loss during training
        norm bool : do normalizatin on the embedding after training
        use_context_weight : give different weights to the nodes in a context window
        async_update : asynchronous training
        """
        super(SkipGramModel, self).__init__()
        self.emb_size = emb_size
        self.emb_dimension = emb_dimension
        self.walk_length = walk_length
        self.window_size = window_size
        self.batch_size = batch_size
        self.only_cpu = only_cpu
        self.only_gpu = only_gpu
        self.mixed_train = mix
        self.neg_weight = neg_weight
        self.negative = negative
        self.lr = lr
        self.lap_norm = lap_norm
        self.fast_neg = fast_neg
        self.record_loss = record_loss
        self.norm = norm
        self.use_context_weight = use_context_weight
        self.async_update = async_update
        self.num_threads = num_threads

        # initialize the device as cpu
        self.device = torch.device("cpu")

        # content embedding
        self.u_embeddings = nn.Embedding(
            self.emb_size, self.emb_dimension, sparse=True)
        # context embedding
        self.v_embeddings = nn.Embedding(
            self.emb_size, self.emb_dimension, sparse=True)
        # initialze embedding
        initrange = 1.0 / self.emb_dimension
        init.uniform_(self.u_embeddings.weight.data, -initrange, initrange)
        init.constant_(self.v_embeddings.weight.data, 0)

        # lookup_table is used for fast sigmoid computing
        self.lookup_table = torch.sigmoid(torch.arange(-6.01, 6.01, 0.01))
        self.lookup_table[0] = 0.
        self.lookup_table[-1] = 1.
        if self.record_loss:
            self.logsigmoid_table = torch.log(torch.sigmoid(torch.arange(-6.01, 6.01, 0.01)))
            self.loss = []

        # indexes to select positive/negative node pairs from batch_walks
        self.index_emb_posu, self.index_emb_posv = init_emb2pos_index(
            self.walk_length,
            self.window_size,
            self.batch_size)
        self.index_emb_negu, self.index_emb_negv = init_emb2neg_index(
            self.walk_length,
            self.window_size,
            self.negative,
            self.batch_size)

        if self.use_context_weight:
            self.context_weight = init_weight(
                self.walk_length,
                self.window_size,
                self.batch_size)

        # adam
        self.state_sum_u = torch.zeros(self.emb_size)
        self.state_sum_v = torch.zeros(self.emb_size)

        # gradients of nodes in batch_walks
        self.grad_u, self.grad_v = init_empty_grad(
            self.emb_dimension,
            self.walk_length,
            self.batch_size)

    def create_async_update(self):
        """ Set up the async update subprocess.
        """
        self.async_q = Queue(1)
        self.async_p = mp.Process(target=async_update, args=(self.num_threads, self, self.async_q))
        self.async_p.start()

    def finish_async_update(self):
        """ Notify the async update subprocess to quit.
        """
        self.async_q.put((None, None, None, None, None))
        self.async_p.join()

    def share_memory(self):
        """ share the parameters across subprocesses """
        self.u_embeddings.weight.share_memory_()
        self.v_embeddings.weight.share_memory_()
        self.state_sum_u.share_memory_()
        self.state_sum_v.share_memory_()

    def set_device(self, gpu_id):
        """ set gpu device """
        self.device = torch.device("cuda:%d" % gpu_id)
        print("The device is", self.device)
        self.lookup_table = self.lookup_table.to(self.device)
        if self.record_loss:
            self.logsigmoid_table = self.logsigmoid_table.to(self.device)
        self.index_emb_posu = self.index_emb_posu.to(self.device)
        self.index_emb_posv = self.index_emb_posv.to(self.device)
        self.index_emb_negu = self.index_emb_negu.to(self.device)
        self.index_emb_negv = self.index_emb_negv.to(self.device)
        self.grad_u = self.grad_u.to(self.device)
        self.grad_v = self.grad_v.to(self.device)
        if self.use_context_weight:
            self.context_weight = self.context_weight.to(self.device)

    def all_to_device(self, gpu_id):
        """ move all of the parameters to a single GPU """
        self.device = torch.device("cuda:%d" % gpu_id)
        self.set_device(gpu_id)
        self.u_embeddings = self.u_embeddings.cuda(gpu_id)
        self.v_embeddings = self.v_embeddings.cuda(gpu_id)
        self.state_sum_u = self.state_sum_u.to(self.device)
        self.state_sum_v = self.state_sum_v.to(self.device)

    def fast_sigmoid(self, score):
        """ do fast sigmoid by looking up in a pre-defined table """
        idx = torch.floor((score + 6.01) / 0.01).long()
        return self.lookup_table[idx]

    def fast_logsigmoid(self, score):
        """ do fast logsigmoid by looking up in a pre-defined table """
        idx = torch.floor((score + 6.01) / 0.01).long()
        return self.logsigmoid_table[idx]

    def fast_learn(self, batch_walks, neg_nodes=None):
        """ Learn a batch of random walks in a fast way. It has the following features:
            1. It calculating the gradients directly without the forward operation.
            2. It does sigmoid by a looking up table.
        Specifically, for each positive/negative node pair (i,j), the updating procedure is as following:
            score = self.fast_sigmoid(u_embedding[i].dot(v_embedding[j]))
            # label = 1 for positive samples; label = 0 for negative samples.
            u_embedding[i] += (label - score) * v_embedding[j]
            v_embedding[i] += (label - score) * u_embedding[j]
        Parameters
        ----------
        batch_walks list : a list of node sequnces
        lr float : current learning rate
        neg_nodes torch.LongTensor : a long tensor of sampled true negative nodes. If neg_nodes is None,
            then do negative sampling randomly from the nodes in batch_walks as an alternative.
        Usage example
        -------------
        batch_walks = [torch.LongTensor([1,2,3,4]),
                       torch.LongTensor([2,3,4,2])])
        lr = 0.01
        neg_nodes = None
        """
        lr = self.lr

        # [batch_size, walk_length]
        if isinstance(batch_walks, list):
            nodes = torch.stack(batch_walks)
        elif isinstance(batch_walks, torch.LongTensor):
            nodes = batch_walks
        if self.only_gpu:
            nodes = nodes.to(self.device)
            if neg_nodes is not None:
                neg_nodes = neg_nodes.to(self.device)
        emb_u = self.u_embeddings(nodes).view(-1, self.emb_dimension).to(self.device)
        emb_v = self.v_embeddings(nodes).view(-1, self.emb_dimension).to(self.device)

        ## Postive
        bs = len(batch_walks)
        if bs < self.batch_size:
            index_emb_posu, index_emb_posv = init_emb2pos_index(
                self.walk_length,
                self.window_size,
                bs)
            index_emb_posu = index_emb_posu.to(self.device)
            index_emb_posv = index_emb_posv.to(self.device)
        else:
            index_emb_posu = self.index_emb_posu
            index_emb_posv = self.index_emb_posv

        # num_pos: the number of positive node pairs generated by a single walk sequence
        # [batch_size * num_pos, dim]
        emb_pos_u = torch.index_select(emb_u, 0, index_emb_posu)
        emb_pos_v = torch.index_select(emb_v, 0, index_emb_posv)

        pos_score = torch.sum(torch.mul(emb_pos_u, emb_pos_v), dim=1)
        pos_score = torch.clamp(pos_score, max=6, min=-6)
        # [batch_size * num_pos, 1]
        score = (1 - self.fast_sigmoid(pos_score)).unsqueeze(1)
        if self.record_loss:
            self.loss.append(torch.mean(self.fast_logsigmoid(pos_score)).item())

        # [batch_size * num_pos, dim]
        if self.lap_norm > 0:
            grad_u_pos = score * emb_pos_v + self.lap_norm * (emb_pos_v - emb_pos_u)
            grad_v_pos = score * emb_pos_u + self.lap_norm * (emb_pos_u - emb_pos_v)
        else:
            grad_u_pos = score * emb_pos_v
            grad_v_pos = score * emb_pos_u

        if self.use_context_weight:
            if bs < self.batch_size:
                context_weight = init_weight(
                    self.walk_length,
                    self.window_size,
                    bs).to(self.device)
            else:
                context_weight = self.context_weight
            grad_u_pos *= context_weight
            grad_v_pos *= context_weight

        # [batch_size * walk_length, dim]
        if bs < self.batch_size:
            grad_u, grad_v = init_empty_grad(
                self.emb_dimension,
                self.walk_length,
                bs)
            grad_u = grad_u.to(self.device)
            grad_v = grad_v.to(self.device)
        else:
            self.grad_u = self.grad_u.to(self.device)
            self.grad_u.zero_()
            self.grad_v = self.grad_v.to(self.device)
            self.grad_v.zero_()
            grad_u = self.grad_u
            grad_v = self.grad_v
        grad_u.index_add_(0, index_emb_posu, grad_u_pos)
        grad_v.index_add_(0, index_emb_posv, grad_v_pos)

        ## Negative
        if bs < self.batch_size:
            index_emb_negu, index_emb_negv = init_emb2neg_index(
                self.walk_length, self.window_size, self.negative, bs)
            index_emb_negu = index_emb_negu.to(self.device)
            index_emb_negv = index_emb_negv.to(self.device)
        else:
            index_emb_negu = self.index_emb_negu
            index_emb_negv = self.index_emb_negv
        emb_neg_u = torch.index_select(emb_u, 0, index_emb_negu)

        if neg_nodes is None:
            emb_neg_v = torch.index_select(emb_v, 0, index_emb_negv)
        else:
            emb_neg_v = self.v_embeddings.weight[neg_nodes].to(self.device)

        # [batch_size * walk_length * negative, dim]
        neg_score = torch.sum(torch.mul(emb_neg_u, emb_neg_v), dim=1)
        neg_score = torch.clamp(neg_score, max=6, min=-6)
        # [batch_size * walk_length * negative, 1]
        score = - self.fast_sigmoid(neg_score).unsqueeze(1)
        if self.record_loss:
            self.loss.append(self.negative * self.neg_weight * torch.mean(self.fast_logsigmoid(-neg_score)).item())

        grad_u_neg = self.neg_weight * score * emb_neg_v
        grad_v_neg = self.neg_weight * score * emb_neg_u

        grad_u.index_add_(0, index_emb_negu, grad_u_neg)
        if neg_nodes is None:
            grad_v.index_add_(0, index_emb_negv, grad_v_neg)

        ## Update
        nodes = nodes.view(-1)

        # use adam optimizer
        grad_u = adam(grad_u, self.state_sum_u, nodes, lr, self.device, self.only_gpu)
        grad_v = adam(grad_v, self.state_sum_v, nodes, lr, self.device, self.only_gpu)
        if neg_nodes is not None:
            grad_v_neg = adam(grad_v_neg, self.state_sum_v, neg_nodes, lr, self.device, self.only_gpu)

        if self.mixed_train:
            grad_u = grad_u.cpu()
            grad_v = grad_v.cpu()
            if neg_nodes is not None:
                grad_v_neg = grad_v_neg.cpu()
            else:
                grad_v_neg = None

            if self.async_update:
                grad_u.share_memory_()
                grad_v.share_memory_()
                nodes.share_memory_()
                if neg_nodes is not None:
                    neg_nodes.share_memory_()
                    grad_v_neg.share_memory_()
                self.async_q.put((grad_u, grad_v, grad_v_neg, nodes, neg_nodes))

        if not self.async_update:
            self.u_embeddings.weight.data.index_add_(0, nodes.view(-1), grad_u)
            self.v_embeddings.weight.data.index_add_(0, nodes.view(-1), grad_v)
            if neg_nodes is not None:
                self.v_embeddings.weight.data.index_add_(0, neg_nodes.view(-1), grad_v_neg)
        return

    def forward(self, pos_u, pos_v, neg_v):
        ''' Do forward and backward. It is designed for future use. '''
        emb_u = self.u_embeddings(pos_u)
        emb_v = self.v_embeddings(pos_v)
        emb_neg_v = self.v_embeddings(neg_v)

        score = torch.sum(torch.mul(emb_u, emb_v), dim=1)
        score = torch.clamp(score, max=6, min=-6)
        score = -F.logsigmoid(score)

        neg_score = torch.bmm(emb_neg_v, emb_u.unsqueeze(2)).squeeze()
        neg_score = torch.clamp(neg_score, max=6, min=-6)
        neg_score = -torch.sum(F.logsigmoid(-neg_score), dim=1)

        # return torch.mean(score + neg_score)
        return torch.sum(score), torch.sum(neg_score)

    def save_embedding(self, dataset, file_name):
        """ Write embedding to local file. Only used when node ids are numbers.
        Parameter
        ---------
        dataset DeepwalkDataset : the dataset
        file_name str : the file name
        """
        embedding = self.u_embeddings.weight.cpu().data.numpy()
        if self.norm:
            embedding /= np.sqrt(np.sum(embedding * embedding, 1)).reshape(-1, 1)
        np.save(file_name, embedding)

    def save_embedding_pt(self, dataset, file_name):
        """ For ogb leaderboard.
        """
        try:
            max_node_id = max(dataset.node2id.keys())
            if max_node_id + 1 != self.emb_size:
                print("WARNING: The node ids are not serial.")

            embedding = torch.zeros(max_node_id + 1, self.emb_dimension)
            index = torch.LongTensor(list(map(lambda id: dataset.id2node[id], list(range(self.emb_size)))))
            embedding.index_add_(0, index, self.u_embeddings.weight.cpu().data)

            if self.norm:
                embedding /= torch.sqrt(torch.sum(embedding.mul(embedding), 1) + 1e-6).unsqueeze(1)
            torch.save(embedding, file_name)
        except:
            self.save_embedding_pt_dgl_graph(dataset, file_name)

    def save_embedding_pt_dgl_graph(self, dataset, file_name):
        """ For ogb leaderboard """
        embedding = torch.zeros_like(self.u_embeddings.weight.cpu().data)
        valid_seeds = torch.LongTensor(dataset.valid_seeds)
        valid_embedding = self.u_embeddings.weight.cpu().data.index_select(0,
                                                                           valid_seeds)
        embedding.index_add_(0, valid_seeds, valid_embedding)

        if self.norm:
            embedding /= torch.sqrt(torch.sum(embedding.mul(embedding), 1) + 1e-6).unsqueeze(1)

        torch.save(embedding, file_name)

    def save_embedding_txt(self, dataset, file_name):
        """ Write embedding to local file. For future use.
        Parameter
        ---------
        dataset DeepwalkDataset : the dataset
        file_name str : the file name
        """
        embedding = self.u_embeddings.weight.cpu().data.numpy()
        if self.norm:
            embedding /= np.sqrt(np.sum(embedding * embedding, 1)).reshape(-1, 1)
        with open(file_name, 'w') as f:
            f.write('%d %d\n' % (self.emb_size, self.emb_dimension))
            for wid in range(self.emb_size):
                e = ' '.join(map(lambda x: str(x), embedding[wid]))
                f.write('%s %s\n' % (str(dataset.id2node[wid]), e))


import os
from dgl.data.utils import download, extract_archive
from dgl.data import DGLDataset
from dgl.data.utils import load_graphs


class AcademicDataset(DGLDataset):

    _prefix = 'https://s3.cn-north-1.amazonaws.com.cn/dgl-data/'
    _urls = {
        'academic4HetGNN': 'dataset/academic4HetGNN.zip',
        'acm4GTN': 'dataset/acm4GTN.zip',
        'acm4NSHE': 'dataset/acm4NSHE.zip',
        'acm4NARS': 'dataset/acm4NARS.zip',
        'acm4HeCo': 'dataset/acm4HeCo.zip',
        'imdb4MAGNN': 'dataset/imdb4MAGNN.zip',
        'imdb4GTN': 'dataset/imdb4GTN.zip',
        'DoubanMovie': 'dataset/DoubanMovie.zip',
        'dblp4MAGNN': 'dataset/dblp4MAGNN.zip',
        'yelp4HeGAN': 'dataset/yelp4HeGAN.zip',
        'yelp4rec': 'dataset/yelp4rec.zip',
        'HNE-PubMed': 'dataset/HNE-PubMed.zip',
        'MTWM': 'dataset/MTWM.zip',
        'amazon4SLICE': 'dataset/amazon4SLICE.zip'
    }

    def __init__(self, name, raw_dir=None, force_reload=False, verbose=True):
        assert name in ['acm4GTN', 'acm4NSHE', 'academic4HetGNN', 'imdb4MAGNN', 'imdb4GTN', 'HNE-PubMed', 'MTWM',
                        'DoubanMovie', 'dblp4MAGNN', 'acm4NARS', 'acm4HeCo', 'yelp4rec', 'yelp4HeGAN', 'amazon4SLICE']
        self.data_path = './openhgnn/' + self._urls[name]
        self.g_path = './openhgnn/dataset/' + name + '/graph.bin'
        raw_dir = './openhgnn/dataset'
        url = self._prefix + self._urls[name]
        super(AcademicDataset, self).__init__(name=name,
                                        url=url,
                                        raw_dir=raw_dir,
                                        force_reload=force_reload,
                                        verbose=verbose)

    def download(self):
        # download raw data to local disk
        # path to store the file
        if os.path.exists(self.data_path):  # pragma: no cover
           pass
        else:
            file_path = os.path.join(self.raw_dir)
            # download file
            download(self.url, path=file_path)
        extract_archive(self.data_path, os.path.join(self.raw_dir, self.name))

    def process(self):
        # process raw data to graphs, labels, splitting masks
        g, _ = load_graphs(self.g_path)
        self._g = g[0]

    def __getitem__(self, idx):
        # get one example by index
        assert idx == 0, "This dataset has only one graph"
        return self._g

    def __len__(self):
        # number of data examples
        return 1

    def save(self):
        # save processed data to directory `self.save_path`
        pass

    def load(self):
        # load processed data from directory `self.save_path`
        pass

    def has_cache(self):
        # check whether there are processed data in `self.save_path`
        pass


import torch as th
import torch.nn as nn
import torch.nn.functional as F


class HeteroEmbedLayer(nn.Module):
    r"""
    Embedding layer for featureless heterograph.

    Parameters
    -----------
    n_nodes_dict : dict[str, int]
        Key of dict means node type,
        value of dict means number of nodes.
    embed_size : int
        Dimension of embedding,
    activation : callable activation function/layer or None, optional
        If not None, applies an activation function to the updated node features.
        Default: ``None``.
    dropout : float, optional
        Dropout rate. Default: ``0.0``
    """

    def __init__(self,
                 n_nodes_dict,
                 embed_size,
                 embed_name='embed',
                 activation=None,
                 dropout=0.0):
        super(HeteroEmbedLayer, self).__init__()

        self.embed_size = embed_size
        self.embed_name = embed_name
        self.activation = activation
        self.dropout = nn.Dropout(dropout)

        # create weight embeddings for each node for each relation
        self.embeds = nn.ParameterDict()
        for ntype, nodes in n_nodes_dict.items():
            embed = nn.Parameter(th.FloatTensor(nodes, self.embed_size))
            # initrange = 1.0 / self.embed_size
            # nn.init.uniform_(embed, -initrange, initrange)
            nn.init.xavier_uniform_(embed, gain=nn.init.calculate_gain('relu'))
            self.embeds[ntype] = embed

    def forward(self, ):
        r"""
        Returns
        -------
        The output embeddings.
        """
        out_feature = {}
        for key, embed in self.embeds.items():
            out_feature[key] = embed
        return out_feature

    def forward_nodes(self, nodes_dict):
        r"""

        Parameters
        ----------
        nodes_dict : dict[str, th.Tensor]
            Key of dict means node type, value of dict means idx of nodes.

        Returns
        -------
        out_feature : dict[str, th.Tensor]
            Output feature.
        """
        out_feature = {}
        for key, nid in nodes_dict.items():
            out_feature[key] = self.embeds[key][nid]
        return out_feature


class multi_Linear(nn.Module):
    def __init__(self, linear_list, bias=False):
        super(multi_Linear, self).__init__()
        self.encoder = nn.ModuleDict({})
        for linear in linear_list:
            self.encoder[linear[0]] = nn.Linear(in_features=linear[1], out_features=linear[2], bias=bias)

    def reset_parameters(self):
        nn.init.xavier_uniform_(self.encoder.weight)

    def forward(self, name_linear, h):
        h = self.encoder[name_linear](h)
        return h


class multi_2Linear(nn.Module):
    def __init__(self, linear_list, bias=False):
        super(multi_2Linear, self).__init__()
        hidden_dim = 16
        self.hidden_layer = nn.ModuleDict({})
        self.output_layer = nn.ModuleDict({})
        for linear in linear_list:
            self.hidden_layer[linear[0]] = nn.Linear(in_features=linear[1], out_features=hidden_dim, bias=bias)
            self.output_layer[linear[0]] = nn.Linear(in_features=hidden_dim, out_features=linear[2], bias=bias)

    def forward(self, name_linear, h):
        h = F.relu(self.hidden_layer[name_linear](h))
        h = self.output_layer[name_linear](h)
        return h


class hetero_linear(nn.Module):
    def __init__(self, linear_list, bias=False):
        super(hetero_linear, self).__init__()
        # In one graph, the node with different node type may have different dimension size as the input.
        # The feature_mapping NN feature the input dimension project to another same dimension size.
        # So the out_dim is just a scalar.

        # n_feats are graph dgl.ndata name.
        self.encoder = multi_Linear(linear_list, bias)

    def forward(self, h_dict):
        h_out = {}
        for ntype, h in h_dict.items():
            h = self.encoder(ntype, h)
            h_out[ntype] = h
        return h_out


import torch
from torch import nn
import torch.nn.functional as F

import dgl
from dgl.nn.pytorch.softmax import edge_softmax
import dgl.function as fn


class AttConv(nn.Module):
    """
    Description
    -----------
    Attention-based convolution was introduced in `Hybrid Micro/Macro Level Convolution for Heterogeneous Graph Learning
    <https://arxiv.org/abs/>`__
    and mathematically is defined as follows:

    """

    def __init__(self, in_feats: tuple, 
                 out_feats: int, 
                 num_heads: int, 
                 dropout: float = 0.0, 
                 negative_slope: float = 0.2):
        """
        Parameters
        ----------
        in_feats : pair of ints
            Input feature size.
        out_feats : int
            Output feature size.
        num_heads : int
            Number of heads in Multi-Head Attention.
        dropout : float, optional
            Dropout rate, defaults: 0.
        negative_slope : float, optional
            Negative slope rate, defaults: 0.2.
        """
        super(AttConv, self).__init__()
        self._in_src_feats, self._in_dst_feats = in_feats[0], in_feats[1]
        self._out_feats = out_feats
        self._num_heads = num_heads

        self.dropout = nn.Dropout(dropout)
        self.leaky_relu = nn.LeakyReLU(negative_slope)

    def forward(self,
                graph: dgl.DGLHeteroGraph,
                feat: tuple,
                dst_node_transformation_weight: nn.Parameter,
                src_node_transformation_weight: nn.Parameter,
                src_nodes_attention_weight: nn.Parameter):
        r"""Compute graph attention network layer.
        Parameters
        ----------
        graph : specific relational DGLHeteroGraph
        feat : pair of torch.Tensor
            The pair contains two tensors of shape (N_{in}, D_{in_{src}})` and (N_{out}, D_{in_{dst}}).
        dst_node_transformation_weight: Parameter (input_dst_dim, n_heads * hidden_dim)
        src_node_transformation_weight: Parameter (input_src_dim, n_heads * hidden_dim)
        src_nodes_attention_weight: Parameter (n_heads, 2 * hidden_dim)
        Returns
        -------
        torch.Tensor, shape (N, H, D_out)` where H is the number of heads, and D_out is size of output feature.
        """
        with graph.local_scope():
            # Tensor, (N_src, input_src_dim)
            feat_src = self.dropout(feat[0])
            # Tensor, (N_dst, input_dst_dim)
            feat_dst = self.dropout(feat[1])
            # Tensor, (N_src, n_heads, hidden_dim) -> (N_src, input_src_dim) * (input_src_dim, n_heads * hidden_dim)
            feat_src = torch.matmul(feat_src, src_node_transformation_weight).view(-1, self._num_heads, self._out_feats)
            # Tensor, (N_dst, n_heads, hidden_dim) -> (N_dst, input_dst_dim) * (input_dst_dim, n_heads * hidden_dim)
            feat_dst = torch.matmul(feat_dst, dst_node_transformation_weight).view(-1, self._num_heads, self._out_feats)

            # first decompose the weight vector into [a_l || a_r], then
            # a^T [Wh_i || Wh_j] = a_l Wh_i + a_r Wh_j, This implementation is much efficient
            # Tensor, (N_dst, n_heads, 1),   (N_dst, n_heads, hidden_dim) * (n_heads, hidden_dim)
            e_dst = (feat_dst * src_nodes_attention_weight[:, :self._out_feats]).sum(dim=-1, keepdim=True)
            # Tensor, (N_src, n_heads, 1),   (N_src, n_heads, hidden_dim) * (n_heads, hidden_dim)
            e_src = (feat_src * src_nodes_attention_weight[:, self._out_feats:]).sum(dim=-1, keepdim=True)
            # (N_src, n_heads, hidden_dim), (N_src, n_heads, 1)
            graph.srcdata.update({'ft': feat_src, 'e_src': e_src})
            # (N_dst, n_heads, 1)
            graph.dstdata.update({'e_dst': e_dst})
            # compute edge attention, e_src and e_dst are a_src * Wh_src and a_dst * Wh_dst respectively.
            graph.apply_edges(fn.u_add_v('e_src', 'e_dst', 'e'))
            # shape (edges_num, heads, 1)
            e = self.leaky_relu(graph.edata.pop('e'))

            # compute softmax
            graph.edata['a'] = edge_softmax(graph, e)

            graph.update_all(fn.u_mul_e('ft', 'a', 'msg'), fn.sum('msg', 'ft'))
            # (N_dst, n_heads * hidden_dim),   (N_dst, n_heads, hidden_dim) reshape
            dst_features = graph.dstdata.pop('ft').reshape(-1, self._num_heads * self._out_feats)

            dst_features = F.relu(dst_features)

        return dst_features


import dgl
import torch as th
import torch.nn as nn
import dgl.function as fn
import torch.nn.functional as F
from . import BaseModel, register_model


@register_model('HetGNN')
class HetGNN(BaseModel):
    r"""
    HetGNN[KDD2019]-
    `Heterogeneous Graph Neural Network <https://dl.acm.org/doi/abs/10.1145/3292500.3330961>`_
    `Source Code Link <https://github.com/chuxuzhang/KDD2019_HetGNN>`_

    The author of the paper only gives the academic dataset.

    Attributes
    -----------
    Het_Aggrate : nn.Module
        Het_Aggregate
    """
    @classmethod
    def build_model_from_args(cls, args, hg):
        return cls(hg, args)

    def __init__(self, hg, args):
        super(HetGNN, self).__init__()
        self.Het_Aggregate = Het_Aggregate(hg.ntypes, args.dim)
        self.ntypes = hg.ntypes
        self.device = args.device

        self.loss_fn = HetGNN.compute_loss

    def forward(self, hg, h=None):
        if h is None:
            h = self.extract_feature(hg, self.ntypes)
        x = self.Het_Aggregate(hg, h)
        return x

    def evaluator(self):
        self.link_preddiction()
        self.node_classification()

    def get_embedding(self):
        input_features = self.model.extract_feature(self.hg, self.hg.ntypes)
        x = self.model(self.model.preprocess(self.hg, self.args).to(self.args.device), input_features)
        return x

    def link_preddiction(self):
        x = self.get_embedding()
        self.model.lp_evaluator(x[self.category].to('cpu').detach(), self.train_batch, self.test_batch)

    def node_classification(self):
        x = self.get_embedding()
        self.model.nc_evaluator(x[self.category].to('cpu').detach(), self.labels, self.train_idx, self.test_idx)

    @staticmethod
    def compute_loss(pos_score, neg_score):
        # an example hinge loss
        loss = []
        for i in pos_score:
            loss.append(F.logsigmoid(pos_score[i]))
            loss.append(F.logsigmoid(-neg_score[i]))
        loss = th.cat(loss)
        return -loss.mean()

    @staticmethod
    def extract_feature(g, ntypes):
        input_features = {}
        for n in ntypes:
            ndata = g.srcnodes[n].data
            data = {}
            data['dw_embedding'] = ndata['dw_embedding']
            data['abstract'] = ndata['abstract']
            if n == 'paper':
                data['title'] = ndata['title']
                data['venue'] = ndata['venue']
                data['author'] = ndata['author']
                data['reference'] = ndata['reference']
            input_features[n] = data

        return input_features

    @staticmethod
    def pred(edge_subgraph, x):
        with edge_subgraph.local_scope():
            edge_subgraph.ndata['x'] = x
            for etype in edge_subgraph.canonical_etypes:
                edge_subgraph.apply_edges(
                    dgl.function.u_dot_v('x', 'x', 'score'), etype=etype)
            return edge_subgraph.edata['score']


class ScorePredictor(nn.Module):
    def forward(self, edge_subgraph, x):
        with edge_subgraph.local_scope():
            edge_subgraph.ndata['x'] = x
            for etype in edge_subgraph.canonical_etypes:
                edge_subgraph.apply_edges(
                    dgl.function.u_dot_v('x', 'x', 'score'), etype=etype)
            return edge_subgraph.edata['score']


class Het_Aggregate(nn.Module):
    r"""
    The whole model of HetGNN

    Attributes
    -----------
    content_rnn : nn.Module
        het_content_encoder
    neigh_rnn : nn.Module
        aggregate_het_neigh
    atten_w : nn.ModuleDict[str, nn.Module]


    """
    def __init__(self, ntypes, dim):
        super(Het_Aggregate, self).__init__()
        # ntypes means nodes type name
        self.ntypes =ntypes
        self.dim = dim

        self.content_rnn = het_content_encoder(dim)
        self.neigh_rnn = aggregate_het_neigh(ntypes, dim)

        self.atten_w = nn.ModuleDict({})
        for n in self.ntypes:
            self.atten_w[n] = nn.Linear(in_features=dim * 2, out_features=1)

        self.softmax = nn.Softmax(dim=1)
        self.activation = nn.LeakyReLU()
        self.drop = nn.Dropout(p=0.5)
        self.bn = nn.BatchNorm1d(dim)
        self.embed_d = dim

    def forward(self, hg, h_dict):
        with hg.local_scope():
            content_h = {}
            for ntype, h in h_dict.items():
                content_h[ntype] = self.content_rnn(h)

            neigh_h = self.neigh_rnn(hg, content_h)
            # the content feature of the dst nodes
            dst_h = {k: v[:hg.number_of_dst_nodes(k)] for k, v in content_h.items()}
            out_h = {}
            for n in self.ntypes:
                d_h = dst_h[n]
                batch_size = d_h.shape[0]
                concat_h = []
                concat_emd = []
                for i in range(len(neigh_h[n])):
                    concat_h.append(th.cat((d_h, neigh_h[n][i]), 1))
                    concat_emd.append(neigh_h[n][i])
                concat_h.append(th.cat((d_h, d_h), 1))
                concat_emd.append(d_h)

                concat_h = th.hstack(concat_h).view(batch_size * (len(self.ntypes) + 1), self.dim *2)
                atten_w = self.activation(self.atten_w[n](concat_h)).view(batch_size, len(self.ntypes) + 1)

                atten_w = self.softmax(atten_w).view(batch_size, 1, 4)

                # weighted combination
                concat_emd = th.hstack(concat_emd).view(batch_size, len(self.ntypes) + 1, self.dim)

                weight_agg_batch = th.bmm(atten_w, concat_emd).view(batch_size, self.dim)
                out_h[n] = weight_agg_batch
            return out_h


class het_content_encoder(nn.Module):
    r"""
    The Encoding Heterogeneous Contents(C2) in the paper
    For a specific node type, encoder different content features with a LSTM.

    In paper, it is (b) NN-1: node heterogeneous contents encoder in figure 2.

    Parameters
    ------------
    dim : int
        input dimension

    Attributes
    ------------
    content_rnn : nn.Module
        nn.LSTM encode different content feature
    """
    def __init__(self, dim):
        super(het_content_encoder, self).__init__()
        self.content_rnn = nn.LSTM(dim, int(dim / 2), 1, batch_first=True, bidirectional=True)
        self.content_rnn.flatten_parameters()
        self.dim = dim

    def forward(self, h_dict):
        r"""

        Parameters
        ----------
        h_dict: dict[str, th.Tensor]
            key means different content feature

        Returns
        -------
        content_h : th.tensor
        """
        concate_embed = []
        for _, h in h_dict.items():
            concate_embed.append(h)
        concate_embed = th.cat(concate_embed, 1)
        concate_embed = concate_embed.view(concate_embed.shape[0], -1, self.dim)
        all_state, last_state = self.content_rnn(concate_embed)
        out_h = th.mean(all_state, 1).squeeze()
        return out_h


class aggregate_het_neigh(nn.Module):
    r"""
    Description
    --------------
    It is a Aggregating Heterogeneous Neighbors(C3)
    Same Type Neighbors Aggregation

    """
    def __init__(self, ntypes, dim):
        super(aggregate_het_neigh, self).__init__()
        self.neigh_rnn = nn.ModuleDict({})
        self.ntypes =ntypes
        for n in ntypes:
            self.neigh_rnn[n] = lstm_aggr(dim)

    def forward(self, hg, inputs):
        with hg.local_scope():
            outputs = {}
            for i in self.ntypes:
                outputs[i] = []
            if isinstance(inputs, tuple) or hg.is_block:
                if isinstance(inputs, tuple):
                    src_inputs, dst_inputs = inputs
                else:
                    src_inputs = inputs
                    dst_inputs = {k: v[:hg.number_of_dst_nodes(k)] for k, v in inputs.items()}

                for stype, etype, dtype in hg.canonical_etypes:
                    rel_graph = hg[stype, etype, dtype]
                    if rel_graph.number_of_edges() == 0:
                        continue
                    if stype not in src_inputs or dtype not in dst_inputs:
                        continue
                    dstdata = self.neigh_rnn[stype](
                        rel_graph,
                        (src_inputs[stype], dst_inputs[dtype]))
                    outputs[dtype].append(dstdata)
            else:
                for stype, etype, dtype in hg.canonical_etypes:
                    rel_graph = hg[stype, etype, dtype]
                    if rel_graph.number_of_edges() == 0:
                        continue
                    if stype not in inputs:
                        continue
                    dstdata = self.neigh_rnn[stype](
                        rel_graph,
                        inputs[stype])
                    outputs[dtype].append(dstdata)
            return outputs


class lstm_aggr(nn.Module):
    r"""
    Aggregate the same neighbors with LSTM
    """

    def __init__(self, dim):
        super(lstm_aggr, self).__init__()
        self.lstm = nn.LSTM(dim, int(dim / 2), 1, batch_first=True, bidirectional=True)

        self.lstm.flatten_parameters()
    def _lstm_reducer(self, nodes):
        m = nodes.mailbox['m']  # (B, L, D)
        batch_size = m.shape[0]
        all_state, last_state = self.lstm(m)
        return {'neigh': th.mean(all_state, 1)}

    def forward(self, g, inputs):
        with g.local_scope():
            if isinstance(inputs, tuple) or g.is_block:
                if isinstance(inputs, tuple):
                    src_inputs, dst_inputs = inputs
                else:
                    src_inputs = inputs
                    dst_inputs = {k: v[:g.number_of_dst_nodes(k)] for k, v in inputs.items()}
                g.srcdata['h'] = src_inputs
                g.update_all(fn.copy_u('h', 'm'), self._lstm_reducer)
                h_neigh = g.dstdata['neigh']
            else:
                g.srcdata['h'] = inputs
                g.update_all(fn.copy_u('h', 'm'), self._lstm_reducer)
                h_neigh = g.dstdata['neigh']
            return h_neigh


# from openhgnn.models.micro_layer.LSTM_conv import LSTMConv
# from openhgnn.models.HeteroGraphConv import HeteroGraphConv
# class HetGNNConv(nn.Module):
#     def __init__(self, graph, ntypes, dim):
#         super(HetGNNConv, self).__init__()
#         # ntypes means nodes type name
#         self.ntypes =ntypes
#         self.dim = dim
#
#         # hetero conv modules
#         self.micro_conv = HeteroGraphConv({
#             etype: LSTMConv(dim=dim)
#             for srctype, etype, dsttype in graph.canonical_etypes
#         })
#
#         # different types aggregation module
#         self.macro_conv = AttConv(in_feats=hidden_dim * n_heads, out_feats=hidden_dim,
#                                                              num_heads=n_heads,
#                                                              dropout=dropout, negative_slope=0.2)
#
#         self.atten_w = nn.ModuleDict({})
#         for n in self.ntypes:
#             self.atten_w[n] = nn.Linear(in_features=dim * 2, out_features=1)
#
#         self.softmax = nn.Softmax(dim=1)
#         self.activation = nn.LeakyReLU()
#         self.drop = nn.Dropout(p=0.5)
#         self.bn = nn.BatchNorm1d(dim)
#         self.embed_d = dim
#
#     def forward(self, hg, h):
#         x = self.Het_Aggrate(hg, h)
#         return x




import dgl
import dgl.function as fn
import numpy as np
import array
import torch
import random
import networkx as nx
import os
import pickle
import networkx as nx

from dgl._ffi.base import DGLError
from networkx.exception import NetworkXException
from torch.serialization import save
from networkx.utils import pairwise

"""
Notice:
1. All the ids come from the whole graph
2. This sampler is irrelevant to train/valid/test 
"""
class SLiCESampler(object):
    def __init__(self,g,g_sample,num_walks_per_node,
        beam_width,max_num_edges,walk_type,path_option,save_path) -> None:
        super().__init__()
        self.g=g_sample
        self.g_all=g
        self.g_networkx=dgl.to_networkx(self.g,edge_attrs=['label'])
        self.node_types=dict()
        for node in g.nodes():
            self.node_types[int(node)]=int(g.ndata['_TYPE'][int(node)])
        self.num_walks_per_node=num_walks_per_node
        self.beam_width=beam_width
        self.walk_type=walk_type#pretrain
        self.path_option=path_option
        self.max_num_edges=max_num_edges
        self.last_context=[]
        self.node_context_dict=dict()
        self.edge_context_dict=dict()
        self.save_path=save_path
        self.pretrain_path=os.path.join(save_path,'pretrain')
        self.finetune_path=os.path.join(save_path,'finetune')
        random.seed(10)
    def get_node_subgraph(self,seed_nodes) -> list:
        """
        Params:
        seed_nodes: the tensor of seed nodes for each subgraph
        Return:
        return a list of sampled node subgraph
        """
        if not os.path.exists(self.save_path):
            os.makedirs(self.save_path)
        if not os.path.exists(self.pretrain_path):
            os.makedirs(self.pretrain_path)
        all_contexts=[]
        for source in seed_nodes:
            source=int(source)
            node_contexts=[]
            if source in self.node_context_dict:
                node_contexts=self.node_context_dict[source]
                all_contexts.extend(node_contexts)
                continue
            for _ in range(self.num_walks_per_node):
                if self.walk_type == "bfs":
                    node_context = self.bfs_beam(source, 
                                self.beam_width, self.max_num_edges)
                elif self.walk_type == "dfs":
                    node_context = self.random_chain(source, self.max_num_edges)
                else:
                    print("Unknown context generation strategy, select bfs/dfs")
                # if no subgraph
                if len(node_context)==0:
                    node_context=self.last_context
                node_subgraph=dgl.node_subgraph(self.g,node_context)
                all_contexts.append(node_subgraph)
                node_contexts.append(node_subgraph)
                self.last_context=node_subgraph
            self.node_context_dict[source]=node_contexts
        return all_contexts#list of subgraph
    def get_edge_subgraph(self,seed_edges) -> list:
        """
        Parameters:
        seed_edges: List[(int,int)] a list of edge src and dst
        """
        if not os.path.exists(self.save_path):
            os.makedirs(self.save_path)
        if not os.path.exists(self.finetune_path):
            os.makedirs(self.finetune_path)
        g_networkx=self.g_networkx
        all_context=[]
        for ii,edge in enumerate(seed_edges):#10399 (6682,51)
            src,dst=edge
            if not self.g_all.has_edges_between(src,dst):   #a false edge
                label=0
            else:
                eid=self.g_all.edge_ids(src,dst)
                label=int(self.g_all.edata['label'][eid])
            if (src,dst) in self.edge_context_dict:
                context,pair=self.edge_context_dict[(src,dst)]
                all_context.extend(context)
            else:
                context=self.get_selective_context(g_networkx,src,dst)
                if len(context) == 0:
                    context = [src,dst]   
                this_context=[]
                this_pair=[]
                if isinstance(context[0],list):#context has many walks
                    for each in context:
                        edge_context=dgl.node_subgraph(self.g,each)
                        this_context.append(edge_context)
                        this_pair.append((src,dst,label))
                else:
                    edge_context=dgl.node_subgraph(self.g,context)
                    this_context.append(edge_context)
                    this_pair.append((src,dst,label))
                self.edge_context_dict[(src,dst)]=(this_context,this_pair)
                all_context.extend(this_context)
        return all_context
    
    def get_random_k_nbrs(self,nodeid,exclude_list,k):
        """
        returns randomly sampled 'k' nbrs for a given node as
        list[node_ids]
        if k == -1 :  returns all neighbours
        """
        #get all neighbors
        all_nbrs=[]
        all_nbrs.extend(dgl.sampling.sample_neighbors(self.g,torch.tensor([nodeid]),-1,
                                                edge_dir='out',prob='label').edges(order='eid')[1].tolist())  
        if exclude_list is None or len(exclude_list) == 0:
            all_nbrs = list(set(all_nbrs))
        else:
            all_nbrs = list(set(x for x in all_nbrs if x not in exclude_list))
        if len(all_nbrs)<k:
            all_nbrs.extend(dgl.sampling.sample_neighbors(self.g,torch.tensor([nodeid]),-1,
                                                edge_dir='in',prob='label').edges(order='eid')[0].tolist())                   
        if exclude_list is None or len(exclude_list) == 0:
            all_nbrs = list(set(all_nbrs))
        else:
            all_nbrs = list(set(x for x in all_nbrs if x not in exclude_list))
        if k >= 0 and len(all_nbrs) >= k:
            return random.sample(all_nbrs, k)
        else:
            return all_nbrs
    
    def bfs_beam(self, source, beam_width, max_num_edges):
        """
        Given a source node, a beam width and depth=hops
        return a subgraph around source node .
        The subgraph is retuned as [(source, target, relation)]
        """
        source=int(source)
        subgraph = [source]
        G = self.g
        
        source_nbrs = self.get_random_k_nbrs(source, exclude_list=[], k=beam_width)
        cnt=0
        for x in source_nbrs:
            if cnt>=max_num_edges:
                break
            subgraph.append(x)
            cnt+=1
        for src_nbr in source_nbrs:
            src_nbr_hop2_cands = self.get_random_k_nbrs(
                src_nbr, exclude_list=[source], k=beam_width
            )
            for x in src_nbr_hop2_cands:
                if cnt>=max_num_edges:
                    break
                subgraph.append(x)
                cnt+=1
        # print(len(subgraph))
        if len(subgraph)==0:
            print("node %d has no subgraph"%source)
        if len(subgraph)>max_num_edges:
            print("node %d has subgraph bigger than max_num_edges"%source)
        return subgraph
    def random_chain(self, source, max_num_edges):
        # generates a random walk of length=max_num_edges, without cycles
        # starting from source
        source=int(source)
        path = [source]
        nodes_so_far = [source]
        G = self.g
        #原代码有bug，会导致采样数量不对
        for _ in range(max_num_edges):
            nbrs = self.get_random_k_nbrs(source, exclude_list=nodes_so_far, k=max_num_edges)
            nbr = -1
            while True:
                cursor=-1
                if len(nbrs) != 0:
                    for each in nbrs:
                        if each not in nodes_so_far:
                            nbr=each
                            break
                if nbr==-1:#如果没有邻居或者所有邻居均已被采样
                    source=nodes_so_far[cursor]#退回上一个点，重新采样
                    cursor-=1
                    nbrs=self.get_random_k_nbrs(source, exclude_list=nodes_so_far, k=1)
                else:
                    break
            path.append(nbr)
            nodes_so_far.append(source)
            source = nbr
        return path
    def get_selective_context(
        self, G, source, target, valid_patterns=None
    ):  # FIXME - removed dangerous default list init in valid_patterns
        """
        option = 1) all  2) pattern  3)shortest
        1) Return all paths connecting source and target of length <=max_seq_len
        2) Return only paths that match given pattern set
            #valid_patterns = set(['1_1_1', '2_2_2', '1_2_1', '2_1_2'])
            (1.1: generates paths connecting source and target of length <= max_seq_len
             1.2 filter paths by pattern
        3) Return shortest path between source and target
        4) Return random path between source and target
        """
        beam_width=self.beam_width
        max_seq_len=self.max_num_edges
        option=self.path_option
        paths = []
        if option == "shortest":
            try:
                path_node_list = nx.bidirectional_shortest_path(G, source, target)
                if len(path_node_list)<=max_seq_len:
                    paths.append([source]+[target]+path_node_list[1:-1])
                else:
                    paths.append([source]+[target]+path_node_list[1:max_seq_len-1])
            except NetworkXException:
                return []

        elif option == "all":
            paths = list(nx.all_simple_paths(G, source, target, cutoff=max_seq_len))
            # print("found all paths", source, target, paths)
        elif option=='pattern':
            paths = list(nx.all_simple_paths(G, source, target, cutoff=max_seq_len))
        elif option == "random":
            paths = self.get_random_paths_between_nodes(
                source, target, beam_width, max_seq_len, currpath=[source]
            )
        elif option == "default":
            paths = self.get_context(source, target, beam_width, max_seq_len)
        return paths
    
    def get_random_paths_between_nodes(
        self, source, target, beam_width, max_seq_len, currpath
    ):
        """
        Generates random path between source and target , by selecting
        'beam_width' number of neighbors and upto length >= max_seq_len
        """
        if len(currpath) > max_seq_len or source == target:
            return []
        all_paths = []
        exclude_list = []
        source_nbrs = self.get_random_k_nbrs(source, exclude_list, beam_width)
        # print("selected source nbrs", source_nbrs)
        if target in source_nbrs:
            # print("Found path ending in target", source_nbrs, target)
            path = [source,target]
            all_paths.append(path)
        for n in source_nbrs:
            if n != target and n not in currpath:
                new_source = n
                new_path = list(currpath)
                new_path.append(new_source)
                nbr_paths = self.get_random_paths_between_nodes(
                    new_source, target, beam_width, max_seq_len, new_path
                )
                new_path = new_path[0:-1]
                for p in nbr_paths:
                    p.insert(0, new_source)
                all_paths.extend(nbr_paths)
        return all_paths
    
    def get_context(self, source, target, beam_width, max_seq_len):
        all_contexts = []
        # for i in range(self.num_walks_per_node):
        source_context = self.bfs_beam(source, beam_width)[0:max_seq_len]
        # target_context = self.bfs_beam(target, beam_width)[0:max_seq_len]
        all_contexts.append(source_context)
        # all_contexts.append(target_context)
        return all_contexts
    def generate_false_edges2(self, positive_edge_list, save_path):
        """
        This method takes in as input a list of positive edges of form
        (relation, u, v, "1")
        and tries to generates double number of false edges (not present in graph),
        such that for each edge in positive_edge_list:
            (relation,u, v', "0"), here u -> v' does not exist in self.G
            (relation,u', v, "0"), here u'-> v does not exist in self.G
        returns : random.shuffle(true_edges + false_edges)
        """
        # collect nodes of different types
        print("Generating false edges...")
        g=self.g
        node_type_to_ids = dict()
        for node_id in g.nodes():
            node_id=int(node_id)
            node_type=int(g.ndata['_TYPE'][node_id])
            if node_type in node_type_to_ids:
                node_type_to_ids[node_type].append(node_id)
            else:
                node_type_to_ids[node_type] = [node_id]

        # generate false edges for every positive example
        false_edges = []
        for source, target in positive_edge_list:
            # generate false edges of type (source, relation, false_target) for every
            # (source, relation, target) in positive_edge_list
            source=int(source)
            target=int(target)
            
            target_type = self.node_types[target]
            false_target = random.sample(node_type_to_ids[target_type], 1)[0]
            if not g.has_edges_between(source,false_target):
                false_edges.append((source, false_target))
            # generate false edges of type (false_source, relation, target) for every
            # (source, relation, target) in positive_edge_list
            source_type = self.node_types[source]
            false_source = random.sample(node_type_to_ids[source_type], 1)[0]
            if not g.has_edges_between(false_source,target):
                false_edges.append((false_source, target))

        # generate false edges of type (false_source, relation, target) for every
        final_edges = positive_edge_list + false_edges
        labels=[1]*len(positive_edge_list)+[0]*len(false_edges)
        (res_edge,res_label)=self.shuffle_edge_label(final_edges,labels)
        with open(save_path,'wb') as f:
            pickle.dump((res_edge,res_label),f)
        return (res_edge,res_label)
    def shuffle_edge_label(self,edges,labels):
        edges_tuple=list(zip(edges,labels))
        random.shuffle(edges_tuple)
        (res_edge,res_label)=zip(*edges_tuple)
        
        return (res_edge,res_label)


import dgl
from .random_walk_sampler import RandomWalkSampler
import torch

hg = dgl.heterograph({
    ('user', 'follow', 'user'): ([0, 1, 1, 2, 3], [1, 2, 3, 0, 0]),
    ('user', 'view', 'item'): ([0, 0, 1, 2, 3, 3], [0, 1, 1, 2, 2, 1]),
    ('item', 'viewed-by', 'user'): ([0, 1, 1, 2, 2, 1], [0, 0, 1, 2, 3, 3])})


def test_get_center_context_negatives():
    rw_walks = 5
    hetero_sampler = RandomWalkSampler(g=hg, metapath=['follow', 'view', 'viewed-by'] * 3,
                                       rw_walks=rw_walks, window_size=3, neg_size=5)
    for i in range(hg.num_nodes('user') * rw_walks):
        print(hetero_sampler.get_center_context_negatives(i))

    metapath = ['view', 'viewed-by']
    for i, elem in enumerate(metapath):
        if i == 0:
            adj = hg.adj(etype=elem)
        else:
            adj = torch.sparse.mm(adj, hg.adj(etype=elem))
    adj = adj.coalesce()
    g = dgl.graph(data=(adj.indices()[0], adj.indices()[1]))
    g.edata['rw_prob'] = adj.values()
    homo_sampler = RandomWalkSampler(g=g, rw_length=10, rw_walks=rw_walks, window_size=3, neg_size=5)

    for i in range(g.num_nodes() * rw_walks):
        print(homo_sampler.get_center_context_negatives(i))



import torch as th
import torch.nn as nn
import dgl.nn as dglnn
import torch.nn.functional as F
from . import BaseModel, register_model
from openhgnn.layers.micro_layer import CompConv
from ..utils.dgl_graph import edata_in_out_mask
from ..utils import get_nodes_dict

'''
Here, we present the implementation details for each task used for evaluation in the paper. 
For all the tasks, we used COMPGCN build on PyTorch geometric framework (Fey & Lenssen, 2019).

Link Prediction: For evaluation, 200-dimensional embeddings for node and relation embeddings are used. 
    For selecting the best model we perform a hyperparameter search using the validation data over the values listed in Table 8. 
    For training link prediction models, we use the standard binary cross entropy loss with label smoothing Dettmers et al. (2018).

Node Classification: Following Schlichtkrull et al. (2017), we use 10% training data as validation for selecting the best model for both the datasets. 
    We restrict the number of hidden units to 32. We use cross-entropy loss for training our model.

For all the experiments, training is done using Adam optimizer (Kingma & Ba, 2014) and Xavier initialization (Glorot & Bengio, 2010) is used for initializing parameters.
'''


@register_model('CompGCN')
class CompGCN(BaseModel):
    """
    The models of the simplified CompGCN, without using basis vector, for a heterogeneous graph.
    """

    @classmethod
    def build_model_from_args(cls, args, hg):
        return cls(args.hidden_dim, args.hidden_dim, args.out_dim,
                   hg.etypes,
                   get_nodes_dict(hg), len(hg.etypes),
                   args.n_layers, comp_fn=args.comp_fn, dropout=args.dropout
                   )

    def __init__(self, in_dim, hid_dim, out_dim, etypes, n_nodes, n_rels, num_layers=2, comp_fn='sub', dropout=0.0,
                 activation=F.relu, batchnorm=True):
        super(CompGCN, self).__init__()
        self.in_dim = in_dim
        self.hid_dim = hid_dim
        self.out_dim = out_dim
        self.rel_names = list(set(etypes))
        self.rel_names.sort()
        self.n_rels = n_rels
        self.n_nodes = n_nodes
        self.num_layer = num_layers
        self.comp_fn = comp_fn
        self.dropout = dropout
        self.activation = activation
        self.batchnorm = batchnorm

        self.layers = nn.ModuleList()

        self.r_embedding = nn.Parameter(th.FloatTensor(self.n_rels + 1, self.in_dim))

        self.layers.append(CompGraphConvLayer(self.in_dim,
                                              self.hid_dim,
                                              self.rel_names,
                                              comp_fn=self.comp_fn,
                                              activation=self.activation,
                                              batchnorm=self.batchnorm,
                                              dropout=self.dropout))

        # Hidden layers with n - 1 CompGraphConv layers
        for i in range(self.num_layer - 2):
            self.layers.append(CompGraphConvLayer(self.hid_dim,
                                                  self.hid_dim,
                                                  self.rel_names,
                                                  comp_fn=self.comp_fn,
                                                  activation=self.activation,
                                                  batchnorm=self.batchnorm,
                                                  dropout=self.dropout))

        # Output layer with the output class
        self.layers.append(CompGraphConvLayer(self.hid_dim,
                                              self.out_dim,
                                              self.rel_names,
                                              comp_fn=self.comp_fn))
        nn.init.xavier_uniform_(self.r_embedding)

    def forward(self, hg, n_feats):

        # For full graph training, directly use the graph
        # Forward of n layers of CompGraphConv
        r_feats = self.r_embedding
        if hasattr(hg, 'ntypes'):
            # full graph training
            for layer in self.layers:
                n_feats, r_feats = layer(hg, n_feats, r_feats)
        else:
            # mini-batch training
            for layer, block in zip(self.layers, hg):
                n_feats, r_feats = layer(block, n_feats, r_feats)

        return n_feats

    def preprocess(self, hg):
        edata_in_out_mask(hg)


class CompGraphConvLayer(nn.Module):
    """One layer of simplified CompGCN."""

    def __init__(self,
                 in_dim,
                 out_dim,
                 rel_names,
                 comp_fn='sub',
                 activation=None,
                 batchnorm=False,
                 dropout=0):
        super(CompGraphConvLayer, self).__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.comp_fn = comp_fn
        self.actvation = activation
        self.batchnorm = batchnorm
        self.rel_names = rel_names
        # define dropout layer
        self.dropout = nn.Dropout(dropout)

        # define batch norm layer
        if self.batchnorm:
            self.bn = nn.BatchNorm1d(out_dim)
        # define weights of 3 node matrices
        self.W_O = nn.Linear(self.in_dim, self.out_dim)
        self.W_I = nn.Linear(self.in_dim, self.out_dim)
        self.W_S = nn.Linear(self.in_dim, self.out_dim)

        # define weights of the 1 relation matrix
        self.W_R = nn.Linear(self.in_dim, self.out_dim)
        self.conv = dglnn.HeteroGraphConv({
            rel: CompConv(comp_fn=comp_fn,
                          norm='right',
                          _allow_zero_in_degree=True)
            for rel in rel_names
        })

    def forward(self, hg, n_in_feats, r_feats):
        """
        Compute one layer of composition transfer for one relation only in a
        homogeneous graph with bidirectional edges.
        """
        with hg.local_scope():
            # Assign values to source nodes. In a homogeneous graph, this is equal to
            # assigning them to all nodes.

            # Assign feature to all edges with the same value, the r_feats.

            wdict = {}
            for i, etype in enumerate(self.rel_names):
                if etype[:4] == 'rev-' or etype[-4:] == '-rev':
                    W = self.W_I
                else:
                    W = self.W_O
                wdict[etype] = {'Linear': W, 'h_e': r_feats[i + 1]}

            if hg.is_block:
                inputs_src = n_in_feats
                inputs_dst = {k: v[:hg.number_of_dst_nodes(k)] for k, v in n_in_feats.items()}
            else:
                inputs_src = inputs_dst = n_in_feats
            outputs = self.conv(hg, inputs_src, mod_kwargs=wdict)

            for n, emd in outputs.items():
                # Step 4: add results of self-loop
                if self.comp_fn == 'sub':
                    h_self = self.W_S(inputs_dst[n] - r_feats[-1])
                elif self.comp_fn == 'mul':
                    h_self = self.W_S(inputs_dst[n] * r_feats[-1])
                elif self.comp_fn == 'ccorr':
                    h_self = self.W_S(inputs_dst[n], r_feats[-1])
                else:
                    raise Exception('Only supports sub, mul, and ccorr')
                h_self.add_(emd)
                # Compute relation output

                # Use batch norm
                if self.batchnorm:
                    if h_self.shape[0] > 1:
                        h_self = self.bn(h_self)

                # Use drop out
                n_out_feats = self.dropout(h_self)
                # Use activation function
                if self.actvation is not None:
                    n_out_feats = self.actvation(n_out_feats)

                outputs[n] = n_out_feats

        r_out_feats = self.W_R(r_feats)
        r_out_feats = self.dropout(r_out_feats)
        if self.actvation is not None:
            r_out_feats = self.actvation(r_out_feats)
        return outputs, r_out_feats



import copy
import dgl
import numpy as np
import torch as th
from tqdm import tqdm
import torch.nn as nn
import torch
import torch.nn.functional as F
from . import BaseFlow, register_flow
from ..tasks import build_task
from ..utils import extract_embed, get_nodes_dict
from collections.abc import Mapping
from ..models import build_model, HeteroEmbedLayer


@register_flow("demo")
class Demo(BaseFlow):
    """Demo flows."""

    def __init__(self, args):
        super(Demo, self).__init__(args)

        self.args = args
        self.model_name = args.model
        self.device = args.device

        self.task = build_task(args)
        self.hg = self.task.get_graph().to(self.device)

        self.model = build_model(self.model_name).build_model_from_args(self.args, self.hg)
        self.model = self.model.to(self.device)
    def preprocess(self):

        return

    def train(self):
        pass

    def _mini_train_step(self,):
        pass

    def loss_calculation(self, positive_graph, negative_graph, embedding):
        pass

    def _full_train_setp(self):
        pass

    def _test_step(self, split=None, logits=None):
        pass


import numpy as np
import torch as th
from sklearn.cluster import KMeans
from sklearn.metrics import normalized_mutual_info_score, adjusted_rand_score
from sklearn.metrics import f1_score, accuracy_score, ndcg_score, roc_auc_score
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LogisticRegression
import sklearn.metrics as Metric
from ogb.nodeproppred import Evaluator


class Evaluator():
    def __init__(self, seed):
        self.seed = seed

    def cluster(self, n, X, Y):
        X = np.array(X)
        Y = np.array(Y)
        Y_pred = KMeans(n, random_state=self.seed).fit(X).predict(X)
        nmi = normalized_mutual_info_score(Y, Y_pred)
        ari = adjusted_rand_score(Y, Y_pred)
        return nmi, ari

    def classification(self, X, Y):
        X_train, X_test, Y_train, Y_test = train_test_split(X, Y, test_size=0.2, random_state=self.seed)
        LR = LogisticRegression(max_iter=10000)
        LR.fit(X_train, Y_train)
        Y_pred = LR.predict(X_test)

        macro_f1, micro_f1 = f1_node_classification(Y_test, Y_pred)
        return micro_f1, macro_f1

    def f1_node_classification(self, y_label, y_pred):
        macro_f1 = f1_score(y_label, y_pred, average='macro')
        micro_f1 = f1_score(y_label, y_pred, average='micro')
        return dict(Macro_f1=macro_f1, Micro_f1=micro_f1)

    def cal_acc(self, y_true, y_pred):
        return accuracy_score(y_true, y_pred)

    def cal_roc_auc(self, y_true, y_pred):
        return roc_auc_score(y_true, y_pred)

    def mrr_(self, embedding, w, train_triplets, valid_triplets, test_triplets, hits=[], eval_bz=100, eval_p='raw'):
        if eval_p == "filtered":
            mrr = calc_filtered_mrr(embedding, w, train_triplets, valid_triplets, test_triplets, hits)
        else:
            mrr = calc_raw_mrr(embedding, w, test_triplets, hits, eval_bz)
        return mrr

    def ndcg(self, y_score, y_true):
        return ndcg_score(y_true, y_score, 10)

    ''''''
    # Used in HetGNN
    def LR_pred(self, train_X, train_Y, test_X):
        LR = LogisticRegression(max_iter=10000)
        LR.fit(train_X, train_Y)
        pred_Y = LR.predict(test_X)
        # AUC_score = Metric.roc_auc_score(test_Y, pred_Y)
        return pred_Y

    def link_prediction(self, train_X, train_Y, test_X, test_Y):
        pred_Y = self.LR_pred(train_X, train_Y, test_X)
        AUC_score = Metric.roc_auc_score(test_Y, pred_Y)
        macro_f1, micro_f1 = f1_node_classification(test_Y, pred_Y)
        return AUC_score, macro_f1, micro_f1
    
    def author_link_prediction(self, x, train_batch, test_batch):
        train_u, train_v, train_Y = train_batch
        test_u, test_v, test_Y = test_batch

        train_X = concat_u_v(x, th.tensor(train_u), th.tensor(train_v))
        test_X = concat_u_v(x, th.tensor(test_u), th.tensor(test_v))
        train_Y = th.tensor(train_Y)
        test_Y = th.tensor(test_Y)
        return self.link_prediction(train_X, train_Y, test_X, test_Y)
    ''''''
    # Given embedding and labels, train_idx and test_idx, training a LR.
    def nc_with_LR(self, emd, labels, train_idx, test_idx):
        Y_train = labels[train_idx]
        Y_test = labels[test_idx]
        LR = LogisticRegression(max_iter=10000)
        X_train = emd[train_idx]
        X_test = emd[test_idx]
        LR.fit(X_train, Y_train)
        Y_pred = LR.predict(X_test)
        macro_f1, micro_f1 = f1_node_classification(Y_test, Y_pred)
        return micro_f1, macro_f1


def perturb_and_get_raw_rank(embedding, w, a, r, b, test_size, batch_size=100):
    """ Perturb one element in the triplets
    """
    n_batch = (test_size + batch_size - 1) // batch_size
    ranks = []
    for idx in range(n_batch):
        # print("batch {} / {}".format(idx, n_batch))
        batch_start = idx * batch_size
        batch_end = min(test_size, (idx + 1) * batch_size)
        batch_a = a[batch_start: batch_end]
        batch_r = r[batch_start: batch_end]
        emb_ar = embedding[batch_a] * w[batch_r]
        emb_ar = emb_ar.transpose(0, 1).unsqueeze(2) # size: D x E x 1
        emb_c = embedding.transpose(0, 1).unsqueeze(1) # size: D x 1 x V
        # out-prod and reduce sum
        out_prod = th.bmm(emb_ar, emb_c) # size D x E x V
        score = th.sum(out_prod, dim=0) # size E x V
        score = th.sigmoid(score)
        target = b[batch_start: batch_end].to(score.device)
        ranks.append(sort_and_rank(score, target))
    return th.cat(ranks)


def sort_and_rank(score, target):
    _, indices = th.sort(score, dim=1, descending=True)
    indices = th.nonzero(indices == target.view(-1, 1), as_tuple=False)
    indices = indices[:, 1].view(-1)
    return indices


# return MRR (raw), and Hits @ (1, 3, 10)
def calc_raw_mrr(embedding, w, test_triplets, hits=[], eval_bz=100):
    with th.no_grad():
        s = test_triplets[:, 0]
        r = test_triplets[:, 1]
        o = test_triplets[:, 2]
        test_size = test_triplets.shape[0]

        # perturb subject
        ranks_s = perturb_and_get_raw_rank(embedding, w, o, r, s, test_size, eval_bz)
        # perturb object
        ranks_o = perturb_and_get_raw_rank(embedding, w, s, r, o, test_size, eval_bz)

        ranks = th.cat([ranks_s, ranks_o])
        ranks += 1 # change to 1-indexed

        mrr = th.mean(1.0 / ranks.float())
        print("MRR (raw): {:.6f}".format(mrr.item()))

        for hit in hits:
            avg_count = th.mean((ranks <= hit).float())
            print("Hits (raw) @ {}: {:.6f}".format(hit, avg_count.item()))
    return mrr.item()


def perturb_s_and_get_filtered_rank(embedding, w, s, r, o, test_size, triplets_to_filter):
    """ Perturb subject in the triplets
    """
    num_entities = embedding.shape[0]
    ranks = []
    for idx in range(test_size):
        if idx % 1000 == 0:
            print("test triplet {} / {}".format(idx, test_size))
        target_s = s[idx]
        target_r = r[idx]
        target_o = o[idx]
        filtered_s = filter_s(triplets_to_filter, target_s, target_r, target_o, num_entities)
        target_s_idx = int((filtered_s == target_s).nonzero())
        emb_s = embedding[filtered_s]
        emb_r = w[str(target_r.item())]
        emb_o = embedding[target_o]
        emb_triplet = emb_s * emb_r * emb_o
        scores = th.sigmoid(th.sum(emb_triplet, dim=1))
        _, indices = th.sort(scores, descending=True)
        rank = int((indices == target_s_idx).nonzero())
        ranks.append(rank)
    return th.LongTensor(ranks)


def perturb_o_and_get_filtered_rank(embedding, w, s, r, o, test_size, triplets_to_filter):
    """ Perturb object in the triplets
    """
    num_entities = embedding.shape[0]
    ranks = []
    for idx in range(test_size):
        if idx % 10000 == 0:
            print("test triplet {} / {}".format(idx, test_size))
        target_s = s[idx]
        target_r = r[idx]
        target_o = o[idx]
        filtered_o = filter_o(triplets_to_filter, target_s, target_r, target_o, num_entities)
        target_o_idx = int((filtered_o == target_o).nonzero())
        emb_s = embedding[target_s]
        emb_r = w[str(target_r.item())]
        emb_o = embedding[filtered_o]
        emb_triplet = emb_s * emb_r * emb_o
        scores = th.sigmoid(th.sum(emb_triplet, dim=1))
        _, indices = th.sort(scores, descending=True)
        rank = int((indices == target_o_idx).nonzero())
        ranks.append(rank)
    return th.LongTensor(ranks)


def filter_o(triplets_to_filter, target_s, target_r, target_o, num_entities):
    target_s, target_r, target_o = int(target_s), int(target_r), int(target_o)
    filtered_o = []
    # Do not filter out the test triplet, since we want to predict on it
    if (target_s, target_r, target_o) in triplets_to_filter:
        triplets_to_filter.remove((target_s, target_r, target_o))
    # Do not consider an object if it is part of a triplet to filter
    for o in range(num_entities):
        if (target_s, target_r, o) not in triplets_to_filter:
            filtered_o.append(o)
    return th.LongTensor(filtered_o)


def filter_s(triplets_to_filter, target_s, target_r, target_o, num_entities):
    target_s, target_r, target_o = int(target_s), int(target_r), int(target_o)
    filtered_s = []
    # Do not filter out the test triplet, since we want to predict on it
    if (target_s, target_r, target_o) in triplets_to_filter:
        triplets_to_filter.remove((target_s, target_r, target_o))
    # Do not consider a subject if it is part of a triplet to filter
    for s in range(num_entities):
        if (s, target_r, target_o) not in triplets_to_filter:
            filtered_s.append(s)
    return th.LongTensor(filtered_s)


def calc_filtered_mrr(embedding, w, train_triplets, valid_triplets, test_triplets, hits=[]):
    with th.no_grad():
        s = test_triplets[:, 0]
        r = test_triplets[:, 1]
        o = test_triplets[:, 2]
        test_size = test_triplets.shape[0]

        triplets_to_filter = th.cat([train_triplets, valid_triplets, test_triplets]).tolist()
        triplets_to_filter = {tuple(triplet) for triplet in triplets_to_filter}
        print('Perturbing subject...')
        ranks_s = perturb_s_and_get_filtered_rank(embedding, w, s, r, o, test_size, triplets_to_filter)
        print('Perturbing object...')
        ranks_o = perturb_o_and_get_filtered_rank(embedding, w, s, r, o, test_size, triplets_to_filter)

        ranks = th.cat([ranks_s, ranks_o])
        ranks += 1 # change to 1-indexed

        mrr = th.mean(1.0 / ranks.float())
        print("MRR (filtered): {:.6f}".format(mrr.item()))

        for hit in hits:
            avg_count = th.mean((ranks <= hit).float())
            print("Hits (filtered) @ {}: {:.6f}".format(hit, avg_count.item()))
    return mrr.item()


def concat_u_v(x, u_idx, v_idx):
    u = x[u_idx]
    v = x[v_idx]
    emd = th.cat((u, v), dim=1)
    return emd


def f1_node_classification(y_label, y_pred):
    macro_f1 = f1_score(y_label, y_pred, average='macro')
    micro_f1 = f1_score(y_label, y_pred, average='micro')
    return macro_f1, micro_f1


def evaluate_(seed, X, Y, n):
    _evaluation = evaluation(seed)
    NMI, ARI = _evaluation.cluster(n, X, Y)
    micro, macro = _evaluation.classification(X, Y)

    print('<Cluster>        NMI = %.4f, ARI = %.4f' % (NMI, ARI))

    print('<Classification>     Micro-F1 = %.4f, Macro-F1 = %.4f' % (micro, macro))


def evaluate(seed, dataset, emd, g):
    if dataset == 'acm1':
        n = 3
        X = emd['paper'].detach().to('cpu')
        Y = g.nodes['paper'].data['label'].to('cpu')
    elif dataset == 'imdb':
        n = 3
        X = emd['movie'].detach().to('cpu')
        Y = g.nodes['movie'].data['label'].to('cpu')
    evaluate_(seed, X, Y, n)


def node_classification(y, node_data, mode):
    if mode not in ['train_mask', 'test_mask', 'valid_mask']:
        ValueError
    mask = node_data[mode]
    idx = th.nonzero(mask, as_tuple=False).squeeze()
    y_label = node_data['label'][idx].to('cpu')
    y_pred = th.argmax(y[idx], dim=1)

    macro_f1, micro_f1 = f1_node_classification(y_label, y_pred)
    return macro_f1, micro_f1


def cal_loss_f1(y, node_data, loss_func, mode):
    if mode not in ['train_mask', 'test_mask', 'valid_mask']:
        ValueError
    mask = node_data[mode]
    idx = th.nonzero(mask, as_tuple=False).squeeze()
    y_label = node_data['labels'][idx]
    y = y[idx]
    y_pred = th.argmax(y, dim=1)
    loss = loss_func(y, y_label)
    macro_f1, micro_f1 = f1_node_classification(y_label.cpu(), y_pred.cpu())
    return loss, macro_f1, micro_f1


def cal_acc(y_pred, y_true):
    return th.sum(y_pred.argmax(dim=1) == y_true).item() / len(y_true)



import numpy as np
import torch as th
import torch.nn as nn
import dgl.function as fn
import torch.nn.functional as F
from . import BaseModel, register_model
import torch.nn.functional as F

@register_model('KGCN')
class KGCN(BaseModel):
    r"""
    Description
    -----------
    This module KGCN was introduced in `KGCN <https://dl.acm.org/doi/10.1145/3308558.3313417>`__.

    It included two parts:

    - Aggregate the entity representation and its neighborhood representation into the entity's embedding.
        The message function is defined as follow:

        :math:`\mathrm{v}_{\mathcal{N}(v)}^{u}=\sum_{e \in \mathcal{N}(v)} \tilde{\pi}_{r_{v, e}}^{u} \mathrm{e}`

        where :math:`\mathrm{e}` is the representation of entity,
        :math:`\tilde{\pi}_{r_{v, e}}^{u}` is the scalar weight on the edge from entity to entity,
        the result :math:`\mathrm{v}_{\mathcal{N}(v)}^{u}` saves message which is passed from neighbor nodes

        There are three types of aggregators.
        Sum aggregator takes the summation of two representation vectors,
        Concat aggregator concatenates the two representation vectors and
        Neighbor aggregator directly takes the neighborhood representation of entity as the output representation

        :math:`a g g_{s u m}=\sigma\left(\mathbf{W} \cdot\left(\mathrm{v}+\mathrm{v}_{\mathcal{S}(v)}^{u}\right)+\mathbf{b}\right)`

        :math:`agg $_{\text {concat }}=\sigma\left(\mathbf{W} \cdot \text{concat}\left(\mathrm{v}, \mathrm{v}_{\mathcal{S}(v)}^{u}\right)+\mathbf{b}\right)$`

        :math:`\text { agg }_{\text {neighbor }}=\sigma\left(\mathrm{W} \cdot \mathrm{v}_{\mathcal{S}(v)}^{u}+\mathrm{b}\right)`

        In the above equations, :math:`\sigma` is the nonlinear function and
        :math:`\mathrm{W}` and :math:`\mathrm{b}` are transformation weight and bias.
        the representation of an item is bound up with its neighbors by aggregation

    - Obtain scores using final entity representation and user representation
        The final entity representation is denoted as :math:`\mathrm{v}^{u}`,
        :math:`\mathrm{v}^{u}` do dot product with user representation :math:`\mathrm{u}`
        can obtain the probability. The math formula for the above function is:

        :math:`$\hat{y}_{u v}=f\left(\mathbf{u}, \mathrm{v}^{u}\right)$`

    Parameters
    ----------
        g : DGLGraph
            A knowledge Graph preserves relationships between entities
        args : Config
            Model's config
    """
    @classmethod
    def build_model_from_args(cls, args, g):
        return cls(g, args)

    def __init__(self, g, args):
        super(KGCN, self).__init__()
        self.g = g
        self.args = args
        self.in_dim = args.in_dim
        self.out_dim = args.out_dim
        self.entity_emb_matrix = nn.Parameter(th.FloatTensor(self.g.num_nodes(), self.in_dim))
        self.relation_emb_matrix = nn.Parameter(th.FloatTensor(args.n_relation, self.in_dim))
        self.user_emb_matrix = nn.Parameter(th.FloatTensor(args.n_user, self.in_dim))

        self.Aggregate = KGCN_Aggregate(args)
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.uniform_(self.entity_emb_matrix, -1, 1)
        nn.init.uniform_(self.relation_emb_matrix, -1, 1)
        nn.init.uniform_(self.user_emb_matrix, -1, 1)

    
    def get_score(self):
        r"""
        Description
        -----------
            Obtain scores using final entity representation and user representation
        Returns
        -------

        """
        self.user_embeddings = self.user_emb_matrix[np.array(self.userList)]
        self.scores = th.sum(self.user_embeddings * self.item_embeddings, dim=1)
        self.scores_normalized = th.sigmoid(self.scores)


    def get_embeddings(self):
        return self.user_emb_matrix, self.entity_emb_matrix, self.relation_emb_matrix

    def forward(self, blocks, inputdata):
        r"""

        Description
        -----------
            Predict the probability between user and entity

        Parameters
        ----------
            blocks : list
                Blocks saves the information of neighbor nodes in each layer
            inputdata : numpy.ndarray
                Inputdata contains the relationship between the user and the entity

        Returns
        -------
            labels : torch.Tensor
                the label between users and entities
            scores : torch.Tensor
                Probability of users clicking on entitys
        """
        self.data = inputdata
        self.blocks = blocks
        self.user_indices = self.data[:,0]
        self.itemlist = self.data[:,1]
        self.labels = self.data[:,2]
        self.item_embeddings, self.userList,self.labelList = self.Aggregate(blocks, inputdata)
        self.get_score()
        self.labels = th.tensor(self.labelList).to(self.args.device)

        return self.labels, self.scores


class KGCN_Aggregate(nn.Module):
    def __init__(self, args):
        super(KGCN_Aggregate, self).__init__()
        self.args = args
        self.in_dim = args.in_dim
        self.out_dim = args.out_dim
        if self.args.aggregate == 'CONCAT':
            self.agg = nn.Linear(self.in_dim*2, self.out_dim)
        else:
            self.agg = nn.Linear(self.in_dim, self.out_dim)

    def aggregate(self):
        self.sub_g.update_all(fn.u_mul_e('embedding', 'weight', 'm'),fn.sum('m', 'ft'))
        self.userList = []
        self.labelList = []
        embeddingList = []
        for i in range(len(self.data)):
            weightIndex = np.where(self.itemlist==int(self.sub_g.dstdata['_ID'][i]))
            if self.args.aggregate == 'SUM':
                embeddingList.append(self.sub_g.dstdata['embedding'][i] + self.sub_g.dstdata['ft'][i][weightIndex]) 
            elif self.args.aggregate == 'CONCAT':
                embeddingList.append(th.cat([self.sub_g.dstdata['embedding'][i], self.sub_g.dstdata['ft'][i][weightIndex].squeeze(0)],dim=-1)) 
            elif self.args.aggregate == 'NEIGHBOR':
                embeddingList.append(self.sub_g.dstdata['embedding'][i])
            self.userList.append(int(self.user_indices[weightIndex]))
            self.labelList.append(int(self.labels[weightIndex]))

        self.sub_g.dstdata['embedding'] = th.stack(embeddingList).squeeze(1)
        output = F.dropout(self.sub_g.dstdata['embedding'],p=0)
        if self.layer+1 == len(self.blocks):
            self.item_embeddings = th.tanh(self.agg(output))
        else:
            self.item_embeddings = th.relu(self.agg(output))
        
    def forward(self,blocks,inputdata):
        r"""

        Description
        -----------
            Aggregate the entity representation and its neighborhood representation

        Parameters
        ----------
            blocks : list
                Blocks saves the information of neighbor nodes in each layer
            inputdata : numpy.ndarray
                Inputdata contains the relationship between the user and the entity

        Returns
        -------
            item_embeddings : torch.Tensor
                items' embeddings after aggregated
            userList : list
                Users corresponding to items
            labelList : list
                Labels corresponding to items
        """
        self.data = inputdata
        self.blocks = blocks
        self.user_indices = self.data[:,0]
        self.itemlist = self.data[:,1]
        self.labels = self.data[:,2]
        for self.layer in range(len(blocks)):
            self.sub_g = blocks[self.layer]
            self.aggregate()

        return self.item_embeddings, self.userList, self.labelList




import dgl
import torch.nn as nn
from . import BaseModel, register_model
from dgl.nn.pytorch.conv import APPNPConv
from ..layers.MetapathConv import MetapathConv
from ..utils.utils import extract_metapaths
from ..layers.macro_layer.SemanticConv import SemanticAttention


@register_model('HPN')
class HPN(BaseModel):
    r"""
    Description
    ------------
    This model shows an example of using dgl.metapath_reachable_graph on the original heterogeneous
    graph.HPN from paper `Heterogeneous Graph Propagation Network
    <https://ieeexplore.ieee.org/abstract/document/9428609>`__.
    The author did not provide codes. So, we complete it according to the implementation of HAN


    .. math::
        \mathbf{Z}^{\Phi}=\mathcal{P}_{\Phi}(\mathbf{X})=g_\Phi(f_\Phi(\mathbf{X}))

    where :math:`\mathbf{X}` denotes initial feature matrix and :math:`\mathbf{Z^\Phi}` denotes semantic-specific node embedding.

    .. math::
        \mathbf{H}^{\Phi}=f_\Phi(\mathbf{X})=\sigma(\mathbf{X} \cdot \mathbf{W}^\Phi+\mathbf{b}^{\Phi})

    where :math:`\mathbf{H}^{\Phi}` is projected node feature matrix

    .. math::
        \mathbf{Z}^{\Phi, k}=g_{\Phi}\left(\mathbf{Z}^{\Phi, k-1}\right)=(1-\gamma) \cdot \mathbf{M}^{\Phi} \cdot \mathbf{Z}^{\Phi, k-1}+\gamma \cdot \mathbf{H}^{\Phi}

    where :math:`\mathbf{Z}^{\Phi,k}` denotes node embeddings learned by k-th layer semantic propagation mechanism. :math:`\gamma` is a weight scalar which indicates the
    importance of characteristic of node in aggregating process.
    We use MetapathConv to finish Semantic Propagation and Semantic Fusion.



    Parameters
    ------------
    meta_paths : list
        contain multiple meta-paths.
    category : str
        The category means the head and tail node of metapaths.
    in_size : int
        input feature dimension.
    out_size : int
        out dimension.
    dropout : float
        Dropout probability.
    k_layer : int
        propagation times.
    alpha : float
        Value of restart probability.
    edge_drop : float, optional
        The dropout rate on edges that controls the
        messages received by each node. Default: ``0``.


    """
    @classmethod
    def build_model_from_args(cls, args, hg):
        if args.meta_paths_dict is None:
            meta_paths = extract_metapaths(args.category, hg.canonical_etypes)
        else:
            meta_paths = args.meta_paths_dict

        return cls(meta_paths=meta_paths, category=args.out_node_type,
                    in_size=args.hidden_dim,
                    out_size=args.out_dim,
                    dropout=args.dropout,
                    k_layer=args.k_layer,
                    alpha=args.alpha,
                    edge_drop=args.edge_drop
                   )

    def __init__(self, meta_paths, category, in_size,  out_size,  dropout, k_layer, alpha, edge_drop):
        super(HPN, self).__init__()
        self.category = category
        self.layers = nn.ModuleList()
        self.layers.append(HPNLayer(meta_paths, in_size,  dropout, k_layer, alpha, edge_drop))
        self.linear = nn.Linear(in_size, out_size)

    def forward(self, g, h_dict):
    
        for gnn in self.layers:
            h_dict = gnn(g, h_dict)
        out_dict = {ntype: self.linear(h_dict[ntype]) for ntype in self.category}
    
        return out_dict


class HPNLayer(nn.Module):

    def __init__(self, meta_paths_dict, in_size, dropout, k_layer, alpha, edge_drop):
        super(HPNLayer, self).__init__()

        # semantic projection function fΦ projects node into semantic space

        self.hidden = nn.Sequential(
            nn.Linear(in_features=in_size, out_features=in_size, bias=True),
            nn.ReLU()
        )
        self.meta_paths_dict = meta_paths_dict

        semantic_attention = SemanticAttention(in_size=in_size)
        mods = nn.ModuleDict({mp: APPNPConv(k_layer, alpha, edge_drop) for mp in meta_paths_dict})
        self.model = MetapathConv(meta_paths_dict, mods, semantic_attention)
        self._cached_graph = None
        self._cached_coalesced_graph = {}

    def forward(self, g, h_dict):
        r"""
        Parameters
        -----------
        g : DGLHeteroGraph
            The heterogeneous graph
        h : tensor
            The input features

        Returns
        --------
        h : tensor
            The output features
        """
        h_dict = {ntype: self.hidden(h_dict[ntype]) for ntype in h_dict}
        if self._cached_graph is None or self._cached_graph is not g:
            self._cached_graph = g
            self._cached_coalesced_graph.clear()
            for mp, mp_value in self.meta_paths_dict.items():
                self._cached_coalesced_graph[mp] = dgl.metapath_reachable_graph(
                        g, mp_value)

        h_dict = self.model(self._cached_coalesced_graph, h_dict)
        return h_dict





import os
import torch
from abc import ABC, abstractmethod

from ..tasks import build_task
from ..layers.HeteroLinear import HeteroFeature
from ..utils import get_nodes_dict


class BaseFlow(ABC):
    candidate_optimizer = {
        'Adam': torch.optim.Adam,
        'SGD': torch.optim.SGD
    }

    def __init__(self, args):
        """

        Parameters
        ----------
        args

        Attributes
        -------------
        evaluate_interval: int
            the interval of evaluation in validation
        """
        super(BaseFlow, self).__init__()
        self.evaluator = None
        self.evaluate_interval = 1
        if hasattr(args, '_checkpoint'):
            self._checkpoint = os.path.join(args._checkpoint,
                                            f"{args.model}_{args.dataset}.pt")
        else:
            if hasattr(args, 'load_from_pretrained'):
                self._checkpoint = os.path.join("./openhgnn/output/{}".format(args.model),
                                                f"{args.model}_{args.dataset}_{args.task}.pt")
            else:
                self._checkpoint = None

        if not hasattr(args, 'HGB_results_path') and args.dataset[:3] == 'HGB':
            args.HGB_results_path = os.path.join("./openhgnn/output/{}/{}_{}.txt".format(args.model, args.dataset[5:], args.seed))

        self.args = args
        self.logger = self.args.logger
        self.model_name = args.model
        self.device = args.device
        self.task = build_task(args)
        self.hg = self.task.get_graph().to(self.device)
        self.args.meta_paths_dict = self.task.dataset.meta_paths_dict
        self.patience = args.patience
        self.max_epoch = args.max_epoch
        self.optimizer = None
        self.loss_fn = self.task.get_loss_fn()

    def preprocess(self):
        r"""
        Every trainerflow should run the preprocess_feature if you want to get a feature preprocessing.
        The Parameters in input_feature will be added into optimizer and input_feature will be added into the model.

        Attributes
        -----------
        input_feature : HeteroFeature
            It will return the processed feature if call it.

        """
        if hasattr(self.args, 'activation'):
            if hasattr(self.args.activation, 'weight'):
                import torch.nn as nn
                act = nn.PReLU()
            else:
                act = self.args.activation
        else:
            act = None
        # useful type selection
        if hasattr(self.args, 'feat'):
            pass
        else:
            # Default 0, nothing to do.
            self.args.feat = 0
        self.feature_preprocess(act)
        self.optimizer.add_param_group({'params': self.input_feature.parameters()})
        # for early stop, load the model with input_feature module.
        self.model.add_module('input_feature', self.input_feature)
        self.load_from_pretrained()

    def feature_preprocess(self, act):
        """
        Feat
            0, 1 ,2
        Node feature
            1 node type & more than 1 node types
            no feature

        Returns
        -------

        """

        if self.hg.ndata.get('h', {}) == {} or self.args.feat == 2:
            if self.hg.ndata.get('h', {}) == {}:
                self.logger.feature_info('Assign embedding as features, because hg.ndata is empty.')
            else:
                self.logger.feature_info('feat2, drop features!')
                self.hg.ndata.pop('h')
            self.input_feature = HeteroFeature({}, get_nodes_dict(self.hg), self.args.hidden_dim,
                                               act=act).to(self.device)
        elif self.args.feat == 0:
            self.input_feature = self.init_feature(act)
        elif self.args.feat == 1:
            if self.args.task != 'node_classification':
                self.logger.feature_info('\'feat 1\' is only for node classification task, set feat 0!')
                self.input_feature = self.init_feature(act)
            else:
                h_dict = self.hg.ndata.pop('h')
                self.logger.feature_info('feat1, preserve target nodes!')
                self.input_feature = HeteroFeature({self.category: h_dict[self.category]}, get_nodes_dict(self.hg), self.args.hidden_dim,
                                                   act=act).to(self.device)

    def init_feature(self, act):
        self.logger.feature_info("Feat is 0, nothing to do!")
        if isinstance(self.hg.ndata['h'], dict):
            # The heterogeneous contains more than one node type.
            input_feature = HeteroFeature(self.hg.ndata['h'], get_nodes_dict(self.hg),
                                               self.args.hidden_dim, act=act).to(self.device)
        elif isinstance(self.hg.ndata['h'], torch.Tensor):
            # The heterogeneous only contains one node type.
            input_feature = HeteroFeature({self.hg.ntypes[0]: self.hg.ndata['h']}, get_nodes_dict(self.hg),
                                               self.args.hidden_dim, act=act).to(self.device)
        return input_feature

    @abstractmethod
    def train(self):
        pass
    
    def _full_train_step(self):
        r"""
        Train with a full_batch graph
        """
        raise NotImplementedError
    
    def _mini_train_step(self):
        r"""
        Train with a mini_batch seed nodes graph
        """
        raise NotImplementedError
    
    def _full_test_step(self):
        r"""
        Test with a full_batch graph
        """
        raise NotImplementedError
    
    def _mini_test_step(self):
        r"""
        Test with a mini_batch seed nodes graph
        """
        raise NotImplementedError

    def load_from_pretrained(self):
        if hasattr(self.args, 'load_from_pretrained') and self.args.load_from_pretrained:
            try:
                ck_pt = torch.load(self._checkpoint)
                self.model.load_state_dict(ck_pt)
                self.logger.info('[Load Model] Load model from pretrained model:' + self._checkpoint)
            except FileNotFoundError:
                self.logger.info('[Load Model] Do not load the model from pretrained, '
                                      '{} doesn\'t exists'.format(self._checkpoint))
        # return self.model

    def save_checkpoint(self):
        if self._checkpoint and hasattr(self.model, "_parameters()"):
            torch.save(self.model.state_dict(), self._checkpoint)



import argparse
import copy
import dgl
import numpy as np
import torch
from tqdm import tqdm
import torch.nn.functional as F
from openhgnn.models import build_model
from . import BaseFlow, register_flow
from ..tasks import build_task
import random
import copy

class GraphSampler2:
    r"""
    A faster sampler than GraphSampler.
    First load graph data to self.hg_dict, then interate.
    """
    def __init__(self, hg, k):
        self.k = k
        self.ets = hg.canonical_etypes
        self.nt_et = {}
        for et in hg.canonical_etypes:
            if et[0] not in self.nt_et:
                self.nt_et[et[0]] = [et]
            else:
                self.nt_et[et[0]].append(et)

        self.hg_dict = {key: {} for key in hg.ntypes}
        for nt in hg.ntypes:
            for nid in range(hg.num_nodes(nt)):
                if nid not in self.hg_dict[nt]:
                    self.hg_dict[nt][nid] = {}
                for et in self.nt_et[nt]:
                    self.hg_dict[nt][nid][et] = hg.successors(nid, et)

    def sample_graph_for_dis(self):
        r"""
        sample three graphs from original graph.

        Note
        ------------
        pos_hg:
            Sampled graph from true graph distribution, that is from the original graph with real node and real relation.
        neg_hg1:
            Sampled graph with true nodes pair but wrong realtion.
        neg_hg2:
            Sampled graph with true scr nodes and realtion but wrong node embedding.
        """
        pos_dict = {}
        neg_dict1 = {}
        neg_dict2 = {}

        for nt in self.hg_dict.keys():
            for src in self.hg_dict[nt].keys():
                for i in range(self.k):
                    et = random.choice(self.nt_et[nt])
                    dst = random.choice(self.hg_dict[nt][src][et])
                    if et not in pos_dict:
                        pos_dict[et] = ([src], [dst])
                    else:
                        pos_dict[et][0].append(src)
                        pos_dict[et][1].append(dst)

                    wrong_et = random.choice(self.ets)
                    while wrong_et == et:
                        wrong_et = random.choice(self.ets)
                    wrong_et = (et[0], wrong_et[1], et[2])

                    if wrong_et not in neg_dict1:
                        neg_dict1[wrong_et] = ([src], [dst])
                    else:
                        neg_dict1[wrong_et][0].append(src)
                        neg_dict1[wrong_et][1].append(dst)

        pos_hg = dgl.heterograph(pos_dict, {nt: len(self.hg_dict[nt].keys()) for nt in self.hg_dict.keys()})
        pos_hg1 = dgl.heterograph(neg_dict1, {nt: len(self.hg_dict[nt].keys()) for nt in self.hg_dict.keys()})
        pos_hg2 = dgl.heterograph(pos_dict, {nt: len(self.hg_dict[nt].keys()) for nt in self.hg_dict.keys()})

        return pos_hg, pos_hg1, pos_hg2

    def sample_graph_for_gen(self):
        d = {}
        for nt in self.hg_dict.keys():
            for src in self.hg_dict[nt].keys():
                for i in range(self.k):
                    et = self.nt_et[nt][random.randint(0, len(self.nt_et[nt]) - 1)]
                    dst = self.hg_dict[nt][src][et][random.randint(0, len(self.hg_dict[nt][src][et]) - 1)]
                    if et not in d:
                        d[et] = ([src], [dst])
                    else:
                        d[et][0].append(src)
                        d[et][1].append(dst)

        return dgl.heterograph(d, {nt: len(self.hg_dict[nt].keys()) for nt in self.hg_dict.keys()})

class GraphSampler:
    def __init__(self, hg, k):
        self.nt_et = {}
        for et in hg.canonical_etypes:
            if et[0] not in self.nt_et:
                self.nt_et[et[0]] = [et]
            else:
                self.nt_et[et[0]].append(et)

        self.k = k
        self.hg = hg

    def sample_graph_for_dis(self):
        pos_dict = {}
        neg_dict1 = {}
        neg_dict2 = {}

        for nt in self.hg.ntypes:
            for src in self.hg.nodes(nt):
                for i in range(self.k):
                    et = self.nt_et[nt][random.randint(0, len(self.nt_et[nt])-1)]
                    dst = self.hg.successors(src, et)[random.randint(0, len(self.hg.successors(src, et))-1)]
                    if et not in pos_dict:
                        pos_dict[et] = ([src], [dst])
                    else:
                        pos_dict[et][0].append(src)
                        pos_dict[et][1].append(dst)

                    ets = copy.deepcopy(self.hg.etypes)
                    ets.remove(et[1])
                    wrong_et = (et[0], random.choice(ets), et[2])
                    if wrong_et not in neg_dict1:
                        neg_dict1[wrong_et] = ([src], [dst])
                    else:
                        neg_dict1[wrong_et][0].append(src)
                        neg_dict1[wrong_et][1].append(dst)

        neg_dict2 = copy.deepcopy(pos_dict)

        pos_hg = dgl.heterograph(pos_dict, {nt: self.hg.number_of_nodes(nt) for nt in self.hg.ntypes})
        pos_hg1 = dgl.heterograph(neg_dict1, {nt: self.hg.number_of_nodes(nt) for nt in self.hg.ntypes})
        pos_hg2 = dgl.heterograph(neg_dict2, {nt: self.hg.number_of_nodes(nt) for nt in self.hg.ntypes})

        return pos_hg, pos_hg1, pos_hg2

    def sample_graph_for_gen(self):
        d = {}
        for nt in self.hg.ntypes:
            for src in self.hg.nodes(nt):
                for i in range(self.k):
                    et = self.nt_et[nt][random.randint(0, len(self.nt_et[nt]) - 1)]  # random edge type
                    dst = self.hg.successors(src, et)[random.randint(0, len(self.hg.successors(src, et)) - 1)]
                    if et not in d:
                        d[et] = ([src], [dst])
                    else:
                        d[et][0].append(src)
                        d[et][1].append(dst)

        return dgl.heterograph(d, {nt: self.hg.number_of_nodes(nt) for nt in self.hg.ntypes})


@register_flow('HeGAN_trainer')
class HeGANTrainer(BaseFlow):
    """Node classification flows.
    Supported Model: HeGAN
    Supported Dataset：yelp
    The task is to classify the nodes of HIN(Heterogeneous Information Network).
    Note: If the output dim is not equal the number of classes, a MLP will follow the gnn model.
    """
    def __init__(self, args):
        super().__init__(args)

        self.num_classes = self.task.dataset.num_classes
        self.category = self.task.dataset.category

        self.hg = self.task.get_graph()
        self.model = build_model(self.model_name).build_model_from_args(self.args, self.hg)
        self.model = self.model.to(self.device)

        self.evaluator = self.task.evaluator.classification
        self.evaluate_interval = 1
        self.loss_fn = torch.nn.BCEWithLogitsLoss(reduction='sum')
        self.optim_dis = torch.optim.Adam(self.model.discriminator.parameters(), lr=args.lr_dis, weight_decay=args.wd_dis)
        self.optim_gen = torch.optim.Adam(self.model.generator.parameters(), lr=args.lr_gen, weight_decay=args.wd_gen)
        self.train_idx, self.val_idx, self.test_idx = self.task.get_idx()
        self.labels = self.task.get_labels().to(self.device)
        self.sampler = GraphSampler(self.hg, self.args.n_sample)
        self.sampler2 = GraphSampler2(self.hg, self.args.n_sample)

    def train(self):
        epoch_iter = tqdm(range(self.args.max_epoch))
        for epoch in epoch_iter:
            if self.args.mini_batch_flag:
                dis_loss, gen_loss = self._mini_train_step()
            else:
                dis_loss, gen_loss = self._full_train_step()

            dis_score, gen_score = self._test_step()

            print(epoch)
            print("discriminator:\n\tloss:{:.4f}\n\tmicro_f1: {:.4f},\n\tmacro_f1: {:.4f}".format(dis_loss, dis_score[0], dis_score[1]))
            print("generator:\n\tloss:{:.4f}\n\tmicro_f1: {:.4f},\n\tmacro_f1: {:.4f}".format(gen_loss, gen_score[0], gen_score[1]))

    def _mini_train_step(self):

        dis_loss, gen_loss = None, None
        return dis_loss, gen_loss

    def _full_train_step(self):
        r"""
        Note
        ----
        pos_loss:
            positive graph loss.
        neg_loss1:
            negative graph loss with wrong realtions.
        neg_loss2:
            negativa graph loss with wrong nodes embedding.
        """
        self.model.train()

        gen_loss = None
        dis_loss = None

        # discriminator step
        for _ in range(self.args.epoch_dis):
            # pos_hg, pos_hg1, pos_hg2 = self.sampler.sample_graph_for_dis()
            pos_hg, pos_hg1, pos_hg2 = self.sampler2.sample_graph_for_dis()
            pos_hg = pos_hg.to(self.device)
            pos_hg1 = pos_hg1.to(self.device)
            pos_hg2 = pos_hg2.to(self.device)
            noise_emb = {
                et: torch.tensor(np.random.normal(0.0, self.args.sigma, (pos_hg2.num_edges(et), self.args.emb_size)).astype('float32')).to(self.device)
                for et in pos_hg2.canonical_etypes
            }

            self.model.generator.assign_node_data(pos_hg2, None)
            self.model.generator.assign_edge_data(pos_hg2, None)
            generate_neighbor_emb = self.model.generator.generate_neighbor_emb(pos_hg2, noise_emb)
            pos_score, neg_score1, neg_score2 = self.model.discriminator(pos_hg, pos_hg1, pos_hg2, generate_neighbor_emb)
            pos_loss = self.loss_fn(pos_score, torch.ones_like(pos_score))
            neg_loss1 = self.loss_fn(neg_score1, torch.zeros_like(neg_score1))
            neg_loss2 = self.loss_fn(neg_score2, torch.zeros_like(neg_score2))
            dis_loss = pos_loss + neg_loss2 + neg_loss1

            self.optim_dis.zero_grad()
            dis_loss.backward()
            self.optim_dis.step()

        # generator step
        dis_node_emb, dis_relation_matrix = self.model.discriminator.get_parameters()
        for _ in range(self.args.epoch_gen):
            # gen_hg = self.sampler.sample_graph_for_gen()
            gen_hg = self.sampler2.sample_graph_for_gen()
            noise_emb = {
                # et: torch.tensor(np.random.normal(0.0, self.args.sigma, self.gen_hg.edata['e'][et].shape))
                et: torch.tensor(np.random.normal(0.0, self.args.sigma, (gen_hg.num_edges(et), self.args.emb_size)).astype('float32')).to(self.device)
                for et in gen_hg.canonical_etypes
            }
            gen_hg = gen_hg.to(self.device)
            score = self.model.generator(gen_hg, dis_node_emb, dis_relation_matrix, noise_emb)
            gen_loss = self.loss_fn(score, torch.ones_like(score))
            self.optim_gen.zero_grad()
            gen_loss.backward()
            self.optim_gen.step()

        return dis_loss.item(), gen_loss.item()

    def _test_step(self, split=None, logits=None):
        self.model.eval()
        self.model.generator.eval()
        self.model.discriminator.eval()

        with torch.no_grad():
            dis_emb = self.model.discriminator.nodes_embedding[self.category]
            gen_emb = self.model.generator.nodes_embedding[self.category]

            dis_metric = self.evaluator(dis_emb.cpu(), self.labels.cpu())
            gen_metric = self.evaluator(gen_emb.cpu(), self.labels.cpu())

            return dis_metric, gen_metric


